#|
************************************
* Graphing interface for Macintosh *
*               CLASP              *
************************************

22-Jul-94  ms    gathered as volume

|#

;;;
;;; graph-interface.lisp
;;;

#|
================================================================
Purpose ========================================================
================================================================
Defines an interface to the plotter.


================================================================
Status =========================================================
================================================================
In-progress.

================================================================
Change history =================================================
================================================================
6-Jun-94  mc	Created.
10-24-94 Added code to the `generate-plots' method of histogram to handle categorical data.  (Westy)
11-28-94 Added code to the `generate-plots' method of histogram to special case small numbers of unique data. (Westy)
01-18-94 Added code for cluster plots.  (Westy)
12-01-95 Fixed to work with New Tables (Schmill)
03-22-96 Added code to allow scatter plots of categorical data. More to come.  (Westy)
09-20-96 Fixed bug in `generate-plots' method for scatter-plot.  (Westy)
10-23-96 Added Code for group Means plot. left -- no help, no f-can-compute function, doesn't relabel categorical X (Schmill)
|#


(in-package "CLASP-INTERFACE")

#| ;; This is loaded elsewhere now

(require "QUICKDRAW")

|#

(defvar *clasp-graphs* '())  ;; stores all of the clasp graphs

(defvar *really-ignore-errors* t)

(defmacro maybe-ignore-errors (&body body)
  `(flet ((do-it () (progn . ,body)))
     (if *really-ignore-errors*
       (ignore-errors (do-it))
       (do-it))))
;;;
;;; Define the graph class.
;;;

(defclass graph (clasp::named-object-mixin)
  ((variable-x
    :initform nil
    :initarg :variable-x
    :type CLASP::VARIABLE
    :accessor variable-graph-x
    :documentation "The variable being graphed.")
   (color-by
    :initform nil
    :initarg :color-by
    :accessor graph-color-by 
    :documentation "The graph to color by.")
   (plot-data
    :initform nil
    :initarg :plot-data
    :type 'cons
    :accessor variable-plot-data
    :documentation "The actual generated plot data for use by the plotter package.")
   (graph-view-size
    :initform nil
    :initarg :graph-view-size
    :type FIXNUM
    :accessor graph-view-size
    :documentation "The size of the window that will be opened on open-click.")
   (graph-view-position
    :initform (make-point 15 (+ *MENUBAR-BOTTOM* 10))
    :initarg :graph-view-position
    :type FIXNUM
    :accessor graph-view-position
    :documentation "The position where the window will go when opened.")
   (graph-log-size
    :initform nil
    :initarg :graph-log-size
    :accessor graph-log-size
    :documentation "The size of the graph that will be logged.")
   (graph-keywords
    :initform nil
    :initarg :graph-keywords
    :type LIST
    :accessor graph-keywords
    :documentation "Keywords to give to plot-graph, like :x-first or whatever.")
   )
  (:documentation "The class of graphs.")
  (:DEFAULT-INITARGS
    :graph-view-size *default-graph-size*
    :graph-log-size *default-log-size*))

(defmethod initialize-instance :after ((graph graph) &key)
  "Generates the plot data & adds to the list of clasp graphs."
    (setf (variable-plot-data graph)
          (generate-plots graph))
    (setq *clasp-graphs* (append *clasp-graphs* (list graph))))

(defclass one-variable-graph (graph)
  ()
  (:documentation "A subclass of graph which does not make any extensions to the
original class."))

(defparameter *suppress-graph-labels?* NIL)

(defmethod initialize-instance :after ((graph one-variable-graph) &key)
  "Sets up the name slot and keywords. Also opens the graph, if needed."
  (with-slots (name variable-x) graph
    (setf (CLASP::name graph) 
          (format nil "~:(~A~) OF ~:(~A~)[~:(~A~)]" 
                  (symbol-name (type-of graph)) 
                  (clasp::name variable-x)
                  (clasp::name (clasp::variable-dataset variable-x))))
    (setf (graph-keywords graph) (append (list (if (typep graph 'histogram)
                                                 :x-msg :y-msg)
                                               (clasp::name variable-x))
                                         (if *suppress-graph-labels?*
                                           NIL
                                           (list :title (clasp::name graph)))
                                         (graph-keywords graph)))
    (if *f-auto-open-graphs*
      (open-clasp-object graph))))

;;; the two variable class is an extension on the basic graph
(defclass two-variable-graph (graph)
  ((variable-y
    :initform nil
    :initarg :variable-y
    :type CLASP::variable
    :accessor variable-graph-y
    :documentation "The variable representing the Y axis of the graph."))
  (:documentation "The subclass of graphs with two variables."))

(defmethod initialize-instance :after ((graph two-variable-graph) &key)
  "Sets up the name slot on the two variable graph and fills in keywords."
  (with-slots (name variable-x variable-y color-by) graph
    (setf (CLASP::name graph) 
          (format nil "~:(~A~) OF ~:(~A~)[~:(~A~)] VS ~:(~A~)[~:(~A~)] ~:(~A~)" 
                  (symbol-name (type-of graph)) 
                  (clasp::name variable-x) (clasp::name (clasp::variable-dataset variable-x))
                  (clasp::name variable-y) (clasp::name (clasp::variable-dataset variable-y))
                  (if (not color-by)
                    ""
                    (format nil "colored by ~A" (clasp::name color-by)))))
    (setf (graph-keywords graph) (append
                                  (list :x-msg (clasp::name variable-x) 
                                        :y-msg (clasp::name variable-y))
                                  (if *suppress-graph-labels?*
                                    NIL
                                    (list :title (clasp::name graph)))
                                  (graph-keywords graph)))
    (if *f-auto-open-graphs*
      (open-clasp-object graph))))

;;; the overlaid-graph class is a graph with empty veriable slots

(defclass overlaid-graph (graph)
  ((overlaid-graph-list
    :initform nil
    :initarg :graphs
    :type LIST
    :accessor graphs-overlaid-graphs
    :documentation "A list of graph instances that make up the overlaid graph."))
   (:documentation "A graph consisting of several graphs overlaid."))

(defmethod initialize-instance :after ((graph overlaid-graph) &key)
  "Fixes up the name slot."
  (with-slots (name overlaid-graph-list) graph
    (let ((result-name "Overlay of "))
      (loop for cg in overlaid-graph-list do
            (setq result-name (concatenate 'string result-name (clasp::name cg) ",")))
      ;; Is this still necessary now that we trim the window-title in 'open-clasp-object'?
      (setf result-name (trim-to-maxlen result-name 128))
      (setf (CLASP::name graph) (string-trim "," result-name)))
    (if *f-auto-open-graphs*
      (open-clasp-object graph))))

(defvar *default-number-of-bins* 10)

;;; these are the actual graphs we'll use
(defclass histogram (one-variable-graph)
  ((number-of-bins :initform *default-number-of-bins*)
   (bin-color :initform *blue-color* :accessor dominant-color))
  (:documentation "The histogram class for CLASP histograms. 
Subclass of graph."))

;; Still need to further differentiate here. Later.
(defclass line-or-symbol-plot-mixin ()
  ()
  (:documentation "If this is mixed in we know this is a not a boxy plot."))

(defclass row-line-plot (line-or-symbol-plot-mixin one-variable-graph)
  ((shift-x
    :initform 0
    :initarg :shift-x
    :type REAL
    :accessor shift-x)
   (line-color :initform *blue-color* :accessor dominant-color))
  (:documentation "The row-line-plot class for CLASP row-line-plots. 
Subclass of graph."))

(defclass scatterplot (line-or-symbol-plot-mixin two-variable-graph)
  ()
  (:documentation "The graph class for CLASP scatterplots. 
Subclass of two-variable-graph."))

(defclass line-plot (line-or-symbol-plot-mixin two-variable-graph)
  ((shift-x
    :initform 0
    :initarg :shift-x
    :type REAL
    :accessor shift-x)
   (line-color :initform *green-color* :accessor dominant-color))
  (:documentation "The graph class for CLASP line plots. 
Subclass of two-variable-graph."))

(defclass cluster-plot (line-or-symbol-plot-mixin two-variable-graph)
  ((color :initform *pink-color* :accessor dominant-color)
   (scale :initform 1)
   (scale-type :initform :linear)
   (size  :initform 1)
   (cluster-size :initform 20)
   (filled :initform nil))
  (:documentation "The graph class for CLASP cluster plots.
Subclass of two-variable-graph."))

;; this one has some slots & methods
(defclass regression-plot (line-or-symbol-plot-mixin graph)
  ((variable-y
    :initform nil
    :initarg :variable-y
    :type CLASP::variable
    :accessor variable-graph-y
    :documentation "The variable representing the Y axis of the graph.")
   (regression-slope :accessor regression-slope) 
   (regression-intercept :accessor regression-intercept))
  (:documentation "The graph class for CLASP regression plots. 
Subclass of two-variable-graph."))

(defmethod calc-regression ((graph regression-plot) &key)
  "computes the slope & intercept of the regression line"
  (with-slots (variable-x variable-y) graph
    (multiple-value-bind (slope intercept garb1 garb2 garb3)
                         (clasp::linear-regression-brief 
                                                         (clasp::simplify variable-y) (clasp::simplify variable-x))
      (declare (ignore garb1 garb2 garb3))
      (setf (regression-slope graph) slope)
      (setf (regression-intercept graph) intercept))))

(defmethod initialize-instance :after ((graph regression-plot) &key)
  "computes the slope & intercept of the regression line"
  (with-slots (variable-x variable-y) graph
    (setf (CLASP::name graph) 
          (format nil "~:(~A~) of ~:(~A~)[~:(~A~)] vs ~:(~A~)[~:(~A~)]" 
                  (symbol-name (type-of graph)) 
                  (clasp::name variable-x) (clasp::name (clasp::variable-dataset variable-x))
                  (clasp::name variable-y) (clasp::name (clasp::variable-dataset variable-y))))
    (setf (graph-keywords graph) (append (list :x-msg (clasp::name variable-x) 
                                               :y-msg (clasp::name variable-y)
                                               :title (clasp::name graph))
                                         (graph-keywords graph)))
    (if *f-auto-open-graphs*
      (open-clasp-object graph))))

;; this is really pretty specific right now. the only possible problem is categorical variables.
(defmethod xy-bounds ((graph graph))
  "Returns the x and y bounds of the graph, as well as whether or not they are custom :
values xmin xmax custom-x-p ymin ymax custom-y-p"
  (let* (marker custom-x custom-y
         (plotlist (cdar (variable-plot-data graph)))
         (keywords (graph-keywords graph))
         (xmin (if (setf marker (search (list :X-FIRST) keywords))
                 (progn (setf custom-x t) (elt keywords (1+ marker)))
                 (maybe-ignore-errors (apply #'min (mapcar #'car plotlist)))))
         (xmax (if (setf marker (search (list :X-LAST) keywords))
                 (progn (setf custom-x t) (elt keywords (1+ marker)))
                 (maybe-ignore-errors (apply #'max (mapcar #'car plotlist)))))
         (ymin (if (setf marker (search (list :Y-FIRST) keywords))
                 (progn (setf custom-y t) (elt keywords (1+ marker)))
                 (maybe-ignore-errors (apply #'min (mapcar #'cadr plotlist)))))
         (ymax (if (setf marker (search (list :Y-LAST) keywords))
                 (progn (setf custom-y t) (elt keywords (1+ marker)))
                 (maybe-ignore-errors (apply #'max (mapcar #'cadr plotlist))))))
    (values xmin xmax custom-x ymin ymax custom-y)))

;; here is the method for overlaid graphs; slightly more complex
(defmethod xy-bounds ((graph overlaid-graph))
  "Returns the x and y bounds of the graph, as well as whether or not they are custom :
values xmin xmax custom-x-p ymin ymax custom-y-p"
  (let* (marker custom-xl custom-xh custom-yl custom-yh
                (keywords (graph-keywords graph))
                (xmin (if (setf marker (search (list :X-FIRST) keywords))
                        (progn (setf custom-xl t) (elt keywords (1+ marker)))
                        most-positive-fixnum))
                (xmax (if (setf marker (search (list :X-LAST) keywords))
                        (progn (setf custom-xh t) (elt keywords (1+ marker)))
                        most-negative-fixnum))
                (ymin (if (setf marker (search (list :Y-FIRST) keywords))
                        (progn (setf custom-yl t) (elt keywords (1+ marker)))
                        most-positive-fixnum))
                (ymax (if (setf marker (search (list :Y-LAST) keywords))
                        (progn (setf custom-yh t) (elt keywords (1+ marker)))
                        most-negative-fixnum)))
    (loop for cg in (graphs-overlaid-graphs graph)
          do (multiple-value-bind (nxmin nxmax cx nymin nymax cy)
                                  (xy-bounds cg)
               (if (and (not custom-xl) (< nxmin xmin))
                 (progn (setf xmin nxmin) (setf custom-xl cx)))
               (if (and (not custom-xh) (> nxmax xmax))
                 (progn (setf xmax nxmax) (setf custom-xh cx)))
               (if (and (not custom-yl) (< nymin ymin))
                 (progn (setf ymin nymin) (setf custom-yl cy)))
               (if (and (not custom-yh) (> nymax ymax))
                 (progn (setf ymax nymax) (setf custom-yh cy)))))
    (values xmin xmax (or custom-xl custom-xh) ymin ymax (or custom-yl custom-yh))))

;;;
;;; Specify graphs are openable.
;;;

(defmethod f-can-open-clasp-object ((object graph))
  t)

(defvar *edit-key-p* 'control-key-p
  " the modifier key for editing graphs ")

(defun find-graph-window (graph)
  (find graph (windows) :test 
	#'(lambda (graph window)
	    (and (typep window 'graph-window)
		 (eql graph (window-graph window))))))

(defmethod open-clasp-object ((graph graph))
  "Tries to find an existing window and expose it. If not, passes
control to make-graph-window."
  ;;
  (if (funcall *edit-key-p*)
    (make-graph-editor graph)
    (let ((graph-window (find-graph-window graph)))
      (if graph-window
        (window-select graph-window)
        (if (variable-plot-data graph)
          (generate-graph-and-window graph)
          (gui-warn "No method for plotting this data."))))))

(defmethod open-clasp-object ((graph overlaid-graph))
  "the difference is, we have to regrab the graphs in case they've been tweaked."
  (setf (variable-plot-data graph) (generate-plots graph))
  (if (funcall *edit-key-p*)
    (let ((edit-window (find-window (concatenate 'string (clasp::name graph) " Settings"))))
      (if edit-window 
        (window-select edit-window)
        (make-instance 'overlaid-graph-editor :graph graph)))
    (let ((graph-window (find-graph-window graph)))
      (if graph-window
        (window-select graph-window)
        (if (variable-plot-data graph)
          (generate-graph-and-window graph)
          (gui-warn "No method for plotting this data."))))))

(defclass graph-window (plotter::plot-window)
  ((graph :initform nil :initarg :graph :accessor window-graph))
  (:documentation "A plot window with a slot for the graph it drew."))

(defmethod window-close :before ((pw graph-window))
  "saves the window size state in the graph with a matching title."
  (let ((right-graph (window-graph pw))
        (graph-key (find-window (concatenate 'string "Key for " (window-title pw)) 'WINDOID)))
    (if right-graph (maybe-ignore-errors (setf (graph-view-size right-graph) (view-size pw))))
    (if right-graph (maybe-ignore-errors (setf (graph-view-position right-graph) (view-position pw))))
    (if graph-key (window-close graph-key))
    ))

(defmethod generate-graph-and-window ((graph graph) &rest args &key
                                  (make-permanent-window t)
                                  (use-color ccl:*color-available*)
                                  (mode :display)
                                  &allow-other-keys)
  "Makes a graph window and eventually should fill it all in.
It is expected that more specific methods will be invoked instead."
  (declare (ignore make-permanent-window))
  (with-cursor *watch-cursor*
    (with-slots (plot-data) graph
      (let ((graph-window (make-instance 'graph-window
                            :graph graph
                            :color-p use-color
                            :window-title (trim-to-maxlen (clasp::name graph) *window-title-maximum-length*)
                            :window-show (if (eq mode :log) nil t)
                            :view-position (graph-view-position graph)
                            :view-size (if (eq mode :log) (graph-log-size graph)
                                           (graph-view-size graph)))))
        (remf args :mode)
        (apply #'plotter::plot-graph 
               graph-window
               plot-data
               (append (graph-keywords graph)
                       args))
        (if (eq mode :log)
          (with-slots (plotter::picture) graph-window
            plotter::picture)
          graph-window)))))

#+IGNORE
(defmethod generate-graph-and-window ((plots list) &rest args
                                      &key (name "Plot Segment")
                                      (use-color ccl:*color-available*) &allow-other-keys)
  (with-cursor *watch-cursor*
    (let ((graph-window (make-instance 'graph-window
                          :graph nil
                          :color-p use-color :window-title name
                          :window-show t :view-size *default-graph-size*)))
      (remf args :name)
      (apply #'plotter::plot-graph
             graph-window
             plots args))))

;; unique plot coloring stuff....
(defun calculate-new-color (old-colors)
  (make-color (random 65534) (random 65534) (random 65534)))

(defun new-color (old-colors)
  (let ((good-ones (mapcar #'cadr *color-table*)))
    (loop for x in old-colors do
          (nremove x good-ones))
    (if good-ones
      (car (last good-ones))
      (calculate-new-color old-colors))))

(defparameter *symbologies* (apply #'vector :+ :x :box :* :circle :diamond :triangle
                              (append (loop for x from 97 repeat 26 collect (code-char x))
                                      (loop for x from 65 repeat 26 collect (code-char x)))))

(defun ensure-unique-colors (plots &optional (off-limits '()) (used-symbols -1))
  (let (cbuffer chbuffer)
    (if (null plots)
      '()
      (cons (if (find (plot-color (car plots)) off-limits)
              (progn
                (setf cbuffer (new-color off-limits))
                (push cbuffer off-limits)
                (incf used-symbols)
                (setf chbuffer (svref *symbologies* used-symbols))
                (install-plot-options (car plots) `(:color ,cbuffer :symbol ,chbuffer)))
              (progn
                (push (plot-color (car plots)) off-limits)
                (car plots)))
            (ensure-unique-colors (cdr plots) off-limits used-symbols)))))

(defmethod generate-plots ((overlaid-graph overlaid-graph))
  "Generates the composite of several plots by consing the plots of the others into
a big list."
  (let (unified-plots)
    (with-slots (overlaid-graph-list) overlaid-graph
      (loop for z in overlaid-graph-list do
           (mapcar #'(lambda (plot) (push plot unified-plots))
                   (variable-plot-data z)))
      (ensure-unique-colors unified-plots))))

(defun categorical-sequence-p (numbers)
  "It would nice to replace this with a check in the variable object of the
type slot, but at the present time I am not sure that slot is updated every place it should be."
  (notevery #'numberp numbers))

(defmethod generate-plots ((histogram histogram))
  "generates the plots used by plot-graph"
  (with-slots (variable-x number-of-bins) histogram
    (let* ((numbers (clasp::simplify variable-x)))
      (if (categorical-sequence-p numbers)
        ;; handle categorical
        (multiple-value-bind (binned-numbers bin-width number-of-buckets x-axis-labels)
                             (plotter::histogram-stuff numbers :just-return-numbers t)
          (declare (ignore bin-width number-of-buckets))
          (setf (graph-keywords histogram) (append (graph-keywords histogram) `(:x-axis-labels ,x-axis-labels)))
          (cons (cons (list :histogram (format nil "Histogram of ~:(~A~)" (name-and-dataset variable-x))
                            `(:color ,(dominant-color histogram)
                              :discrete t))
                      binned-numbers)
                '()))
        ;; handle numerical
        (let ((count (length numbers)))
          (multiple-value-bind (min max)
	    (loop for cnt from 0 to (1- count)
		  minimize (elt numbers cnt) into min
		  maximize (elt numbers cnt) into max
		  finally (return (values min max)))
            (let* ((unique-count (length (clasp::extract-unique-values numbers)))
                   ;; use this for cases where we have a small number of distinct points
                   (small-and-unique (and (< count number-of-bins) (= count unique-count)))
                   ;; make sure that the minimum bucket-width is 1 even if the range of numbers is 
                   ;; less than 'number-of-bins'
                   (discrete (and 
                              (every #'integerp numbers)
                              (<= unique-count number-of-bins)
                              ))
                   (bucket-width (cond (small-and-unique nil)
                                       (;(and 
                                        discrete
                                        ;    (<= (- max min) number-of-bins))
                                        1)
                                       (t nil)))
                   (number-of-buckets (cond (small-and-unique nil)
                                            (bucket-width nil)
                                            (t (min number-of-bins count)))))
              (multiple-value-bind (binned-numbers bin-width)
                                   (plotter::histogram-stuff 
                                    numbers 
                                    :bucket-center (when discrete min)
                                    :just-return-numbers t
                                    :bucket-width bucket-width
                                    :number-of-buckets number-of-buckets)
                (cons (cons (list :histogram (format nil "Histogram of ~:(~A~)" (name-and-dataset variable-x))
                                  `(:color ,(dominant-color histogram) 
                                           :border t 
                                           :width ,(if small-and-unique 0 bin-width)
                                           :discrete ,discrete))
                            binned-numbers)
                      '())))))))))

#+OLD
(defmethod generate-plots ((histogram histogram))
  "generates the plots used by plot-graph"
  (with-slots (variable-x number-of-bins) histogram
    (let* ((numbers (clasp::simplify variable-x))
           (count (length numbers)))
      ;; handle categorical
      (if (categorical-variable-p variable-x) ;(notevery #'numberp numbers)
        (multiple-value-bind (binned-numbers bin-width number-of-buckets x-axis-labels)
                             (plotter::histogram-stuff numbers :just-return-numbers t)
          (declare (ignore bin-width number-of-buckets))
          (setf (graph-keywords histogram) (append (graph-keywords histogram) `(:x-axis-labels ,x-axis-labels)))
          (cons (cons (list :histogram (format nil "Histogram of ~:(~A~)" (name-and-dataset variable-x))
                            `(:color ,ccl:*purple-color*))
                      binned-numbers)
                '()))
        (multiple-value-bind (min max)   (loop for cnt from 0 to (1- count)
                                               minimize (elt numbers cnt) into min
                                               maximize (elt numbers cnt) into max
                                               finally (return (values min max)))
          (let* ((unique-count (length (remove-duplicates numbers :test #'=)))
                 ;; make sure that the minimum bucket-width is 1 even if the range of numbers is 
                 ;; less than 'number-of-bins'
                 (discrete (and 
                            (every #'integerp numbers)
                            (<= unique-count number-of-bins)
                            ))
                 (bucket-width (when (and discrete
                                          (<= (- max min) number-of-bins))
                                 1)))
            (multiple-value-bind (binned-numbers bin-width)
                                 (plotter::histogram-stuff 
                                  numbers 
                                  :bucket-center (when discrete min)
                                  :just-return-numbers t
                                  :bucket-width bucket-width
                                  :number-of-buckets (unless bucket-width (min number-of-bins count)))
              (cons (cons (list :histogram (format nil "Histogram of ~:(~A~)" (name-and-dataset variable-x))
                                `(:color ,ccl:*purple-color* 
                                         :border t 
                                         :width ,bin-width
                                         :discrete ,discrete))
                          binned-numbers)
                    '()))))))))

(defmethod generate-plots ((scatterplot scatterplot))
  "generates the plots used by plot-graph for a scatterplot"
  (let ((symbol-index -1))
    (with-slots (variable-x variable-y color-by) scatterplot
      (if color-by
        (let (datasets plots used-colors)
          (setf datasets 
                #-OLD-VARS
                (clasp::partition-on (clasp::variable-dataset variable-x)
                                              (list color-by)
                                              (list variable-x variable-y color-by))
                #+OLD-VARS
                (clasp::partition-dataset (clasp::variable-dataset variable-x)
                                          `(.on. ,(clasp::name color-by))
                                          (list variable-x variable-y color-by))
                )
          (setf plots (mapcan #'(lambda (dset)
                                  ;; This color-by stuff does not coexist well with the categorical recoding code used in here.
                                  ;; Need to fix it.
                                  (incf symbol-index)
                                  (generate-plot-list scatterplot :scatter 
                                                      (get-variable (name variable-x) dset)  
                                                      (get-variable (name variable-y) dset)
                                                      (car (push (new-color used-colors) used-colors))
                                                      (svref *symbologies* symbol-index))
                                  #+OLD
                                  (let ((x (clasp::get-variable (clasp::name variable-x) dset))
                                        (y (clasp::get-variable (clasp::name variable-y) dset)))
                                    (incf symbol-index)
                                    (cons (list :scatter (format nil "Scatter ~:(~A~) on ~:(~A~)" (name-and-dataset x) (name-and-dataset y))
                                                (append `(:color ,(car (push (new-color used-colors) used-colors)))
                                                        (if #+WORKS ccl::*color-available*
                                                            #-WORKS NIL   ;; there is no indicator (yuck)
                                                          NIL
                                                          `(:symbol ,(svref *symbologies* symbol-index)))))
                                          (mapcar #'list (clasp::simplify x) (clasp::simplify y)))))
                              datasets))
          (dolist (dataset datasets)
            (clasp::delete-dataset dataset))
          plots)
        ;; Otherwise
        
        (generate-plot-list scatterplot :SCATTER variable-x variable-y
                            ccl:*blue-color* 
                            :+)
        #+OLD
        (cons (cons (list :scatter (format nil "~:(~A~) on ~:(~A~)" (name-and-dataset variable-x) (name-and-dataset variable-y))
                          `(:color ,ccl:*blue-color* 
                            :symbol :+
                                   ))
                    (mapcar #'cons (clasp::simplify variable-x) 
                            (mapcar #'list (clasp::simplify variable-y))))
              '())))))

(defparameter *max-x-axis-label-length* 15)
(defparameter *max-y-axis-label-length* 25)
(defparameter *max-categorical-values-displayed* 60)

(defmethod generate-plot-list ((self two-variable-graph) 
                               type x-variable y-variable
                               color symbol)
  (let* (;(categorical-x (clasp::categorical-variable-p x-variable))
        ;(categorical-y (clasp::categorical-variable-p y-variable))
        (xs (clasp::simplify x-variable))
        (ys (clasp::simplify y-variable))
        (categorical-x (categorical-sequence-p xs))
        (categorical-y (categorical-sequence-p ys))
        x-labels y-labels)

    (when categorical-x
      (multiple-value-setq (xs x-labels) (plotter::recode-categorical xs))) 
    (when categorical-y
      (multiple-value-setq (ys y-labels) (plotter::recode-categorical ys)))

    ;; Convert alists into plists that the grapher can handle...
    (setf x-labels
          (mapcar #'(lambda (item)
                      `(,(trim-to-maxlen (first item) *max-x-axis-label-length*) ,(rest item)))
                  x-labels))

    (setf y-labels
          (mapcar #'(lambda (item)
                      `(,(trim-to-maxlen (first item) *max-y-axis-label-length*) ,(rest item)))
                  y-labels))
    
    (when (and x-labels (not (> (length x-labels) *max-categorical-values-displayed*)))
      (setf (graph-keywords self) (append (graph-keywords self) `(:x-axis-labels ,x-labels))))
    (when (and y-labels (not (> (length y-labels) *max-categorical-values-displayed*)))
      (setf (graph-keywords self) (append (graph-keywords self) `(:y-axis-labels ,y-labels))))
    (values
     `(((,type ,(concatenate 'string
                             (string-capitalize (name-and-dataset x-variable))
                             " on "
                             (string-capitalize (name-and-dataset y-variable)))
               (:color ,color
                       :symbol ,symbol))
        ,@(mapcar #'list xs ys))))))
  
(defmethod generate-plots ((line-plot line-plot))
  "generates the plots used by plot-graph for a line-plot"
  (with-slots (variable-x variable-y shift-x) line-plot
    (cons (cons (list :polyline-without-vertices (format nil "~:(~A~) on ~:(~A~)" (name-and-dataset variable-x) (name-and-dataset variable-y))
                      `(:color ,(dominant-color line-plot)))
                (mapcar #'(lambda (x y) (list (+ shift-x x) y))
                        (clasp::simplify variable-x)
                        (clasp::simplify variable-y)))
          '())))

(defmethod generate-plots ((self cluster-plot))
    "Generates the plots used by plot-graph for a cluster-plot"
      (with-slots (variable-x variable-y color-by shift-x 
                              size scale scale-type cluster-size filled) self
		  (if color-by
		      (let (datasets
			    plots
			    (used-colors nil))
			(setf datasets
			      (clasp::partition-on (clasp::variable-dataset variable-x)
						   (list color-by)
						   (list variable-x variable-y color-by)))
			(setf plots
			      (mapcar
			       #'(lambda (dset)
				   (let ((x (clasp::get-variable (clasp::name variable-x) dset))
					 (y (clasp::get-variable (clasp::name variable-y) dset)))
				     (cons (list :bubble
						 (format nil "Cluster ~:(~A~) on ~:(~A~)"
							 (name-and-dataset x) (name-and-dataset y))
						 `(:color ,(car (push (new-color used-colors)
								      used-colors))
						   :SIZE ,size 
                                                   :SCALE (,scale-type ,scale)
                                                   :GROUP-SIZE ,cluster-size 
                                                   :FILLED ,filled))
					   (mapcar #'list (clasp::simplify x) (clasp::simplify y)))))
			       datasets))
			(dolist (dataset datasets)
			  (clasp::delete-dataset dataset))
			(values plots))
		      (cons (cons (list :BUBBLE
					(format nil
						"~:(~A~) on ~:(~A~)"
						(name-and-dataset variable-x)
						(name-and-dataset variable-y))
					`(:COLOR ,ccl:*pink-color*
                                          :SIZE ,size 
                                          :SCALE (,scale-type ,scale)
                                          :GROUP-SIZE ,cluster-size 
                                          :FILLED ,filled       
                                          ))
				  (mapcar #'cons (clasp::simplify variable-x)
					  (mapcar #'list (clasp::simplify variable-y))))
			    '()))))

(defmethod generate-plots ((regression-plot regression-plot))
  "generates the plots used by plot-graph for a regression-plot.
this actual method only produces the scatterplot. the regression
line is produced when the window is opened based on the axes."
  (calc-regression regression-plot)
  (let ((symbol-index -1) (ccol))
    (with-slots (variable-x variable-y color-by) regression-plot
      (if color-by
        (let (datasets plots used-colors)
          (setf datasets 
                #-OLD-VARS
                (clasp::partition-on (clasp::variable-dataset variable-x)
                                              (list color-by)
                                              (list variable-x variable-y color-by))
                #+OLD-VARS
                (clasp::partition-dataset (clasp::variable-dataset variable-x)
                                          `(.on. ,(clasp::name color-by))
                                          (list variable-x variable-y color-by))
                )
          (progn
            (loop for dset in datasets do
                  (push (let ((x (clasp::get-variable (clasp::name variable-x) dset))
                              (y (clasp::get-variable (clasp::name variable-y) dset)))
                          (incf symbol-index)
                          (cons (list :scatter (format nil "Scatter ~:(~A~) on ~:(~A~)" (name-and-dataset x) (name-and-dataset y))
                                      (append `(:color ,(setf ccol (car (push (new-color used-colors) used-colors))))
                                              (if #+WORKS ccl::*color-available*
                                                  #-WORKS NIL   ;; there is no indicator (yuck)
                                                  NIL
                                                  `(:symbol ,(svref *symbologies* symbol-index)))))
                                (mapcar #'list (clasp::simplify x) (clasp::simplify y))))
                        plots)
                  (multiple-value-bind (slope int b1 b2 b3)
                                       (clasp::linear-regression-brief (clasp::simplify (clasp::get-variable (clasp::name variable-y) dset))
                                                                       (clasp::simplify (clasp::get-variable (clasp::name variable-x) dset)))
                    (declare (ignore b1 b2 b3))
                    (push (cons (list :slope-intercept-line
                                      (format nil "Regression Line<~,3F int/~,3F slope> ~:(~A~) on ~:(~A~)"
                                              int slope
                                              (name-and-dataset variable-x) (name-and-dataset variable-y))
                                      `(:color ,ccol))
                                (list (list slope int)))
                          plots))
                  (clasp::delete-dataset dset))
            plots))
        ;; not colored-by
        (list (cons (list :scatter (format nil "Regression Scatter ~:(~A~) on ~:(~A~)" (name-and-dataset variable-x) (name-and-dataset variable-y))
                      `(:color ,ccl:*orange-color*))
                (mapcar #'cons (clasp::simplify variable-x) 
                        (mapcar #'list (clasp::simplify variable-y))))
          (cons (list :slope-intercept-line 
                      (format nil "Regression Line<~,3F int/~,3F slope> ~:(~A~) on ~:(~A~)"
                              (regression-intercept regression-plot) (regression-slope regression-plot)
                              (name-and-dataset variable-x) (name-and-dataset variable-y))
                      `(:color ,ccl:*orange-color*))
                (list (list (regression-slope regression-plot) (regression-intercept regression-plot)))))
        ))))

#|
  (with-slots (variable-x variable-y) regression-plot
    (list (cons (list :scatter (format nil "Regression Scatter ~:(~A~) on ~:(~A~)" (name-and-dataset variable-x) (name-and-dataset variable-y))
                      `(:color ,ccl:*orange-color*))
                (mapcar #'cons (clasp::simplify variable-x) 
                        (mapcar #'list (clasp::simplify variable-y))))
          (cons (list :slope-intercept-line 
                      (format nil "Regression Line<~,3F int/~,3F slope> ~:(~A~) on ~:(~A~)"
                              (regression-intercept regression-plot) (regression-slope regression-plot)
                              (name-and-dataset variable-x) (name-and-dataset variable-y))
                      `(:color ,ccl:*orange-color*))
                (list (list (regression-slope regression-plot) (regression-intercept regression-plot)))))))
|#

(defmethod generate-plots ((row-line-plot row-line-plot))
  "generates the plots used by plot-graph for a row-line-plot"
  ;; note the confusing terminology here -- variable-x is a slot of graph,
  ;;  but we really want it to be the y varaible in a row-line-plot
  (with-slots (variable-x) row-line-plot
    (let* ((y (clasp::simplify variable-x))
           (x ;;(if (typep variable-x 'column)
              ;;  (clasp::simplify (clasp::get-variable 'row-number 
              ;;                                        (clasp::variable-dataset variable-x)))
                (loop for row from 1 to (length y)
                      collect (+ (shift-x row-line-plot) row)))
           )
      (cons (cons (list :polyline-without-vertices (format nil "Row Number vs ~:(~A~)" (name-and-dataset variable-x))
                        `(:color ,(dominant-color row-line-plot)))
                  (mapcar #'cons x
                          (mapcar #'list y)))
            '()))))
           

;;;
;;; overlaid graphs temporary launcher
;;;



(defun create-overlaid-graph (graphs)
  "creates an overlaid graph from the selected graphs."
  (if graphs
    (run-clasp-command (make-clasp-command `(make-instance
                                              'overlaid-graph
                                              :graphs
                                              ',graphs
                                              )
                                           :description
                                           (list (list 'command 'overlay-graphs)
                                                 (append (list 'graphs) graphs)))
                       :str-message "Overlaying graphs...."
                       :f-update-dataset-view t)
    (message-dialog "First select some graphs in the caddy or the history, and then choose Overlay Graphs."
                    :ok-text "Oh yeah")))

#|
================================================================
Graphs==========================================================
================================================================
dialog box for Graphs

14-Jul-94 ms	Created.
22-Jul-94 ms    Beta finish.

|#
;;;
;;; Define graph-ov-dialog .
;;;

(defclass graph-ov-dialog (dialog statistic-dialog-mixin)
  ((graph-type :initarg :graph-type)
   (graph-type-name :initarg :graph-type-name))
  (:documentation "A dialog for getting data for the one variable
Graph functions.")
  (:default-initargs
    :WINDOW-TYPE :DOCUMENT
    :WINDOW-TITLE "Graphs"
    :VIEW-SIZE #@(300 70)
    :VIEW-POSITION (best-spot 300 70)
    :VIEW-FONT '("Chicago" 12 :SRCOR :PLAIN)
    :VIEW-SUBVIEWS
    (LIST (make-instance 'variable-pop-up-menu #-CCL-3 :dialog-item-text #+CCL-3 :menu-title "Plot"
                         :view-nick-name 'x-var-item
                         :view-position #@(15 10)
                         :view-size #@(200 20))         ;FIX: 20 is hardcoded
          (make-instance 'compute-button-di :dialog-item-text "Graph"
                         :view-position #@(170 40)))))

(defmethod initialize-instance :after ((dialog graph-ov-dialog) &key (categorical-ok t) &allow-other-keys)
  (setf (allow-categorical-p (view-named 'x-var-item dialog)) categorical-ok)
  (with-slots (graph-type) dialog
    (if (eq graph-type 'row-line-plot)
      (add-subviews dialog
                    (make-instance 'static-text-dialog-item :dialog-item-text "X Shift"
                                   :view-position #@(15 40))
                    (make-instance 'number-entry-item :view-nick-name 'x-shift
                                   :dialog-item-text "0"
                                   :view-position #@(75 40) :view-size #@(59 15))))))

(defmethod f-can-compute-statistics-dialog ((graph-ov-dialog graph-ov-dialog))
  "We can compute if there is at least one sample in the x-var-stdi."
  ;;
  (let* ((x (l-variables-var-pum (view-named 'x-var-item
                                            graph-ov-dialog)))
         (f-can-compute x)
         (str-why-cant-compute (cond (f-can-compute "")
                                     ((null x)
                                      "Need an x variable")
                                     (t "secret reason not to work."))))
    (values f-can-compute
            str-why-cant-compute)))


(defmethod compute-statistics-dialog ((graph-ov-dialog graph-ov-dialog))
  (with-slots (graph-type graph-type-name)  graph-ov-dialog
    (let* ((x (l-variables-var-pum (view-named 'x-var-item
                                               graph-ov-dialog)))
           (xsh (maybe-ignore-errors (number-item-value (view-named 'x-shift graph-ov-dialog)))))
      (loop for this-x in x do
            (graph-ov-variable this-x graph-type graph-type-name xsh)))))


(defmethod graph-ov-variable ((x clasp::variable) graph-type name xsh)
  "Calls function on the arguments, saving
results in the history."
  (let ((clasp-command (make-clasp-command
                        (append
                         `(make-instance
                            ',graph-type
                            :variable-x ,x)
                         (if xsh (list :shift-x xsh)))
                        :description
                        (list (list 'command graph-type)
                              (list 'x-args x)))))
    (run-clasp-command clasp-command
                       :str-message (format nil "Generating ~A..." name)
                       :f-update-dataset-view t)))


(defun graph-ov-dialog-make (graph-type &optional (name (string-capitalize (string graph-type)))
                                        &key (help "No documentation on this graph type.")
                                        (categorical-ok t))
  "Returns a new dialog instance for editing the function."
  ;;
  (let ((selected-vars (items-in-dock 'x-var-dock :select-type :select-or-all)))
    (if selected-vars
      (loop for v in selected-vars do
            (graph-ov-variable v graph-type name nil))
      (make-instance 'graph-ov-dialog
        :graph-type graph-type 
        :graph-type-name name
        :help help
        :window-title name
        :categorical-ok categorical-ok
        ))))

;;; dramatic pause

(defclass overlaid-graph-dialog (dialog statistic-dialog-mixin)
  ()
  (:documentation "A SIMPLE dialog for overlaying graphs")
  (:DEFAULT-INITARGS
    :view-size #@(200 100)
    :view-position (best-spot 200 100)
    :window-title "Overlay Graphs"
    :help '*overlay-graphs-dialog*
    :window-type :document
    :view-subviews (list
                    (make-instance 'compute-button-di
                      :dialog-item-text "overlay"
                      :view-position #@(70 70))
                    (make-instance 'radio-button-dialog-item
                      :dialog-item-text "selected palette graphs"
                      :view-position #@(15 10)
                      :view-nick-name 'use-caddy-rdi)
                    (make-instance 'radio-button-dialog-item
                      :dialog-item-text "selected history graphs"
                      :radio-button-pushed-p (not (find-window "Graph Palette"))
                      :view-position #@(12 30)
                      :view-nick-name 'use-history-rdi))))

(defmethod f-can-compute-statistics-dialog ((ogd overlaid-graph-dialog))
  (let* ((use-caddy (view-named 'use-caddy-rdi ogd))
         (source (if (radio-button-pushed-p use-caddy)
                   (find-window "Graph Palette")
                   (clasp-window-front)))
         (can-compute source)
         (str-compute (if (not source)
                        (if (radio-button-pushed-p use-caddy)
                          "The palette isn't open"
                          "There is no CLASP window open")
                        "CLASP is a dog.")))
    (values can-compute str-compute)))

(defmethod compute-statistics-dialog ((ogd overlaid-graph-dialog))
  (let* ((use-caddy (view-named 'use-caddy-rdi ogd))
         (graphs (if (radio-button-pushed-p use-caddy)
                   (l-caddy-selected-items (find-window "Graph Palette"))
                   (remove-if-not #'(lambda (item) (typep item 'graph))
                                  (l-history-selected-cw (clasp-window-front))))))
    (if graphs
      (create-overlaid-graph graphs)
      (gui-warn "No selected graphs."))))
         
(defun overlay-selected-graphs()
  (let* ((gc (find-window "Graph Palette"))
         (selected-graphs (if gc (l-caddy-selected-items gc))))
    (create-overlaid-graph 
     (if selected-graphs 
       selected-graphs
       (remove-if-not #'(lambda (item) (typep item 'graph))
                      (l-history-selected-cw (clasp-window-front)))))))
#|
================================================================
Graphs==========================================================
================================================================
dialog box for Graphs with two variables

14-Jul-94 ms	Created.
22-Jul-94 ms    Beta finish

|#
;;;
;;; Define graph-tv-dialog .
;;;

(defclass graph-tv-dialog (dialog statistic-dialog-mixin)
  ((graph-type :initarg :graph-type)
   (graph-type-name :initarg :graph-type-name))
  (:documentation "A dialog for getting data for the two variable
Graph functions.")
  (:default-initargs
    :WINDOW-TYPE :DOCUMENT
    :WINDOW-TITLE "Graphs"
    :VIEW-SIZE #@(300 100)
    :VIEW-POSITION (best-spot 300 100)
    :VIEW-FONT '("Chicago" 12 :SRCOR :PLAIN)
    :VIEW-SUBVIEWS
    (LIST (make-instance 'variable-pop-up-menu #-CCL-3 :dialog-item-text #+CCL-3 :menu-title "X(s)"
                         :view-nick-name 'x-var-item
                         :view-position #@(15 5)
                         :view-size #@(200 20))         ;FIX: 20 is hardcoded
          (make-instance 'variable-pop-up-menu #-CCL-3 :dialog-item-text #+CCL-3 :menu-title "Y(s)"
                         :view-nick-name 'y-var-item
                         :view-position #@(15 27)
                         :view-size #@(200 20))         ;FIX: 20 is hardcoded
          (make-instance 'compute-button-di :dialog-item-text "Graph"
                         :view-position #@(140 75)))))

(defmethod initialize-instance :after ((graph-tv-dialog graph-tv-dialog) 
                                       &key (color-by-p nil) (categorical-x t) (categorical-y t)
                                       &allow-other-keys)
  (setf (allow-categorical-p (view-named 'x-var-item graph-tv-dialog)) categorical-x)
  (setf (allow-categorical-p (view-named 'y-var-item graph-tv-dialog)) categorical-y)
  (if color-by-p
    (add-subviews graph-tv-dialog 
                  (make-instance 'variable-stdi :view-nick-name 'color-by-item
                                 :view-size #@(200 20) :view-position #@(75 52))
                  (make-instance 'static-text-dialog-item :dialog-item-text "Color by"
                                :view-position #@(10 52)))))

(defmethod f-can-compute-statistics-dialog ((graph-tv-dialog graph-tv-dialog))
  "We can compute if there is at least one sample in the x-var-stdi."
  ;;
  (let* ((x (l-variables-var-pum (view-named 'x-var-item
                                             graph-tv-dialog)))
         (y (l-variables-var-pum (view-named 'y-var-item
                                             graph-tv-dialog)))
         (cb (maybe-ignore-errors (variable-vset-stdi (view-named 'color-by-item graph-tv-dialog))))
         (f-can-compute (and x y (if cb (and (= 1 (length 
                                                   (remove-duplicates
                                                    (append (mapcar #'clasp::variable-dataset x)
                                                            (mapcar #'clasp::variable-dataset y)))))
                                             (null (remove  (clasp::variable-dataset cb) 
                                                            (mapcar #'clasp::variable-dataset y))))
                                     t)))
         (str-why-cant-compute (cond (f-can-compute "")
                                     ((null x)
                                      "Need an x variable")
                                     ((null y)
                                      "Need a y variable")
                                     (t "All variables must be from the same dataset."))))
    (values f-can-compute
            str-why-cant-compute)))

(defmethod compute-statistics-dialog ((graph-tv-dialog graph-tv-dialog))
  (with-slots (graph-type graph-type-name)  graph-tv-dialog
    (let* ((x (l-variables-var-pum (view-named 'x-var-item
                                                     graph-tv-dialog)))
           (y (l-variables-var-pum (view-named 'y-var-item
                                                     graph-tv-dialog)))
           (cb (maybe-ignore-errors (variable-vset-stdi (view-named 'color-by-item graph-tv-dialog)))))
      (if (> (length y) 1)
        (loop for this-x in x 
              for this-y in y do
              (graph-tv-variables this-x this-y graph-type graph-type-name cb))
        (loop for this-x in x do
              (graph-tv-variables this-x (car y) graph-type graph-type-name cb))))))


(defmethod graph-tv-variables ((x clasp::variable) (y clasp::variable) graph-type name color-by)
  "Calls function on the arguments, saving
results in the history."
  (let ((clasp-command (make-clasp-command
                        `(make-instance
                           ',graph-type
                          :variable-x ,x
                          :variable-y ,y
                          :color-by ,color-by)
                        :description
                        (list (list 'command graph-type)
                              (list 'x-args x)
                              (list 'y-args y)
                              ))))
    (run-clasp-command clasp-command
                       :str-message (format nil "Generating ~A..." name)
                       :f-update-dataset-view t)))


(defun graph-tv-dialog-make (graph-type &optional (name (string-capitalize (string graph-type))) 
                                        (color-by-ok nil)
                                        &key (help "No help for this type of graph.")
                                        (categorical-x t) (categorical-y t))
  "Returns a new dialog instance for editing the function."
  ;;
  (make-instance 'graph-tv-dialog
    :graph-type graph-type 
    :graph-type-name name
    :window-title name
    :help help
    :color-by-p color-by-ok
    :categorical-x categorical-x
    :categorical-y categorical-y
    ))

;;;
;;; plotting tables of data
;;;

;;; the class
(defclass table-plot (line-or-symbol-plot-mixin graph)
  ((table
    :initform nil
    :initarg :table
    :type graphable-table-superclass
    :accessor table-plot-table 
    :documentation "The table which is being plotted."))
  (:documentation "The graph class for plotting table data."))

;;; setup the slots
(defmethod initialize-instance :after ((table-plot table-plot) &key)
  (setf (clasp::name table-plot) 
        (concatenate 'string "Plot of " 
                     (prin1-to-string (table-plot-table table-plot))))
        ; 5/4/95 WAS PRODUCING <BOGUS OBJECT> ERRORS (Westy)
        ;(format nil "Plot of ~A" (clasp::name (table-plot-table table-plot))))
  (make-x-axis-labels table-plot)
  (pushnew (clasp::name table-plot) (graph-keywords table-plot))
  (pushnew :title (graph-keywords table-plot))
  (if *f-auto-open-graphs* (open-clasp-object table-plot)))

;;; generate the x-axis-labels -- hands off to the table object (code is in new-tables.lisp)
(defmethod make-x-axis-labels ((table-plot table-plot))
  (let ((x-labels (x-labels (table-plot-table table-plot))))
    (if x-labels
      (progn
        (remf (graph-keywords table-plot) :x-axis-labels)
        (push x-labels (graph-keywords table-plot))
        (push :x-axis-labels (graph-keywords table-plot))))))

(defmethod generate-plots ((table-plot table-plot))
  "dispatches to the generate-plots method for the table (sneaky). (code is in new-tables.lisp)"
  (generate-plots (table-plot-table table-plot)))

;;; the menu-accessed function

(defun make-table-plot (window)
  (let ((the-table (determine-table window)))
    (if the-table
      (if (generate-plots the-table)
        (run-clasp-command
         (make-clasp-command `(make-instance 'table-plot :table ,the-table)
                             :description (list (list 'command 'plot-table)
                                                (list 'table (clasp::name the-table))))
         :str-message "Plotting table-data...."
         :f-update-dataset-view t)
        (gui-warn "Cannot plot this type of table.")))))

(defmethod determine-table ((window table-view-window))
  (table-view-table (view-named 'table-array-item window)))

;;
;; plot-groups stuff $$$
;;

(defclass graph-group-graph (graph)
  ((parent :initform NIL :initarg :parent :accessor group-parent))
  (:documentation "The graph type that is generated when one plots a group from a graph."))

(defmethod generate-plots ((ggg graph-group-graph))
  (variable-plot-data ggg))

;;; --------------------------------------------------------------------------------
;;; we'll want to open an editor if we c-dc the plot window
(defmethod view-click-event-handler ((window graph-window) (where fixnum))
  (if (and (control-key-p) (double-click-p))
    (open-clasp-object (window-graph window))
    (call-next-method)))

;;; and a little ribbon action too

(defmethod view-activate-event-handler :after ((window graph-window))
  (ribbon-display (window-title window)))

;;; --------------------------------------------------------------------------------

#+OLD
(defun make-histogram (list)
  "Makes a histogram straight from the data. Performs no binning."
  (let ((histo-table (make-hash-table))
        (legal-vals '())
        (bush-league '()) ;; a comically named buffer variable
        (result '()))
    (loop while list do
          (if (setq bush-league (gethash (car list) histo-table))
            (setf (gethash (car list) histo-table) (1+ bush-league))
            (progn
              (setf (gethash (car list) histo-table) 1)
              (setq legal-vals (cons (car list) legal-vals))))
          (setq list (cdr list)))
    (loop while legal-vals do
          (setq result (cons (list (car legal-vals) 
                                   (gethash (car legal-vals) histo-table))
                             result))
          (setq legal-vals (cdr legal-vals)))
    result))

;;; --------------------------------------------------------------------------------
;;; Added 10/23/96 (MS) Group Means Plots

(defclass confidence-mixin ()
  ((confidence :initform NIL :initarg :confidence :accessor confidence)))

(defclass group-means-plot (line-or-symbol-plot-mixin confidence-mixin two-variable-graph)
  ())

(defclass 2way-group-means-plot (line-or-symbol-plot-mixin confidence-mixin two-variable-graph)
  ((group :initform NIL :initarg :group-by :type CLASP::variable
          :accessor variable-graph-y)))

(defparameter *gmp-workspace* (make-hash-table))

(defmethod generate-plots ((gmp group-means-plot))
  (let ((groups '())
        (cat? NIL))
    (clrhash *gmp-workspace*)
    (with-slots (variable-x variable-y confidence) gmp
      (loop for x in (variable-value variable-x)
            for y in (variable-value variable-y) do
            (push y (gethash x *gmp-workspace*)))
      (maphash #'(lambda (k e)
                   (when (not (numberp k))
                     (setq cat? t))
                   (push (cons k e) groups))
               *gmp-workspace*)
      (setq groups (sort groups (if cat? #'(lambda (t1 t2) (string< (princ-to-string t1) (princ-to-string t2)))
                                    #'<) :key #'car))
      (list (cons (list (if confidence :polyline-with-confidence-intervals :polyline-with-vertices) 
                        (format nil "Group Means of ~A by ~A"
                                (string-capitalize (clasp::name variable-y))
                                (string-capitalize (clasp::name variable-x)))
                        (append (list :color (new-color NIL))
                                `(:symbol ,(svref *symbologies* 0))
                                (if confidence `(:vertices? t) '())))
                  (if confidence
                    (mapcar #'(lambda (g) (cons (car g) (multiple-value-list (confidence-interval-t (cdr g) confidence)))) groups)
                    (mapcar #'(lambda (g) (list (car g) (mean (cdr g)))) groups)))))))

(defmethod generate-plots ((gmp 2way-group-means-plot))
  (let ((g-groups '()) (used-colors NIL)
        (cat? NIL))
    (clrhash *gmp-workspace*)
    (with-slots (variable-x variable-y group confidence) gmp
      ;; this pass seperates things into constituent X groups and associates a G tag
      (loop for x in (variable-value variable-x)
            for y in (variable-value variable-y)
            for g in (variable-value group) do
            (when (not (numberp x))
              (setq cat? t))
            (push (cons x y) (gethash g *gmp-workspace*)))
      (maphash #'(lambda (k e)
                   (push (cons k e) g-groups))
               *gmp-workspace*)
      ;; loop over G values
      (loop for g-group in g-groups append
            (let ((x-groups NIL) (color (new-color used-colors)))
              ;;do some setting up
              (push color used-colors)
              (clrhash *gmp-workspace*)
              ;; loop over xy pairs and stuff y's by x
              (loop for xy-pair in (cdr g-group) do
                    (push (cdr xy-pair) (gethash (car xy-pair) *gmp-workspace*)))
              ;; now seperate the x groups
              (maphash #'(lambda (k e)
                           (when (not (numberp k))
                             (setq cat? t))
                           (push (cons k e) x-groups))
                       *gmp-workspace*)
              ;; now sort by x
              (setq x-groups (sort x-groups (if cat? #'(lambda (t1 t2) (string< (princ-to-string t1) (princ-to-string t2)))
                                            #'<) :key #'car))
              ;; now generate the plot group
              (list (cons (list (if confidence :polyline-with-confidence-intervals :polyline-with-vertices) 
                                (format nil "Group Means of ~A by ~A where ~A=~S"
                                        (string-capitalize (clasp::name variable-y))
                                        (string-capitalize (clasp::name variable-x))
                                        (string-capitalize (clasp::name group))
                                        (car g-group))
                                (append (list :color color)
                                        `(:symbol ,(svref *symbologies* 0))
                                        (if confidence `(:vertices? t) '())))
                          (if confidence
                            (mapcar #'(lambda (s) (cons (car s) (multiple-value-list (confidence-interval-t (cdr s) confidence)))) x-groups)
                            (mapcar #'(lambda (s) (list (car s) (mean (cdr s)))) x-groups)))))))))

(defclass graph-ci-dialog (dialog statistic-dialog-mixin) ()
  (:documentation "For doing confidence intervals.")
  (:default-initargs
    :WINDOW-TYPE :DOCUMENT
    :WINDOW-TITLE "Group Means Plot"
    :VIEW-SIZE #@(300 120)
    :VIEW-POSITION (best-spot 300 120)
    :VIEW-FONT '("Chicago" 12 :SRCOR :PLAIN)
    :VIEW-SUBVIEWS
    (LIST (make-instance 'variable-pop-up-menu #-CCL-3 :dialog-item-text #+CCL-3 :menu-title "X(s)"
                         :view-nick-name 'x-var-item
                         :view-position #@(15 5)
                         :view-size #@(200 20))         ;FIX: 20 is hardcoded
          (make-instance 'variable-pop-up-menu #-CCL-3 :dialog-item-text #+CCL-3 :menu-title "Y(s)"
                         :view-nick-name 'y-var-item
                         :view-position #@(15 27)
                         :view-size #@(200 20))         ;FIX: 20 is hardcoded
          (make-instance 'variable-pop-up-menu #-CCL-3 :dialog-item-text #+CCL-3 :menu-title "Group"
                         :view-nick-name 'z-var-item
                         :view-position #@(15 49)
                         :view-size #@(200 20))
          (MAKE-instance 'disabler-checkbox-item :view-position #@(9 85) :view-size #@(92 20)
                            :dialog-item-text "Confidence" :watches-over '(confidence-value)
                            :VIEW-NICK-NAME 'confi-enable-cbdi
                            :check-box-checked-p t)
          (make-instance 'number-entry-item :view-position #@(112 87) :view-size #@(40 16)
                         :allow-negatives NIL :allow-floats t
                         :view-nick-name 'confidence-value :dialog-item-text "0.95")
          (make-instance 'compute-button-di :dialog-item-text "Graph"
                         :view-position #@(235 85)))))

(defmethod f-can-compute-statistics-dialog ((graph-ci-dialog graph-ci-dialog))
  t)

(defmethod compute-statistics-dialog ((gcd graph-ci-dialog))
  (let ((xs (l-variables-var-pum (view-named 'x-var-item gcd)))
        (ys (l-variables-var-pum (view-named 'y-var-item gcd)))
        (gs (l-variables-var-pum (view-named 'z-var-item gcd)))
        (cv (and (check-box-checked-p (view-named 'confi-enable-cbdi gcd)) 
                 (number-item-value (view-named 'confidence-value gcd)))))
    (if (> (length ys) 1)
        (loop for this-x in xs 
              for this-y in ys do
              (graph-means this-x this-y (car gs) cv))
        (loop for this-x in xs do
              (graph-means this-x (car ys) (car gs) cv)))))

(defun graph-means (x y z cv)
  (let ((clasp-command (make-clasp-command
                        `(make-instance
                           ',(if z '2way-group-means-plot 'group-means-plot)
                           :variable-x ,x
                           :variable-y ,y
                           ,@(when z (list :group-by z))
                           ,@(when cv (list :confidence cv)))
                        :description
                        (nconc (list (list 'command 'plot-means)
                                     (list 'x-args x)
                                     (list 'y-args y))
                               (when z (list (list 'z-args z)))
                               (when cv (list (list 'confidence cv)))))))
    (run-clasp-command clasp-command
                       :str-message "Generating means plot..."
                       :f-update-dataset-view t)))

;;; --------------------------------------------------------------------------------
;;; EOF
