;;;
;;; logger.lisp
;;;

#|
Logging facility for Mac CLASP.
Provides ability to log the CLASP Notebook to different destinations.

Currently : preferences are (:test :file :event)

22-Aug-94 ms   created
9-Sep-94  ms   cut and paste functionality added
|#

(in-package "CLASP-INTERFACE")

;;;
;;; TEST LOGGING

(defun open-test-log ()
  (values t "Test logging ok."))

;;;
;;; FILE LOGGING DEFINITIONS

(defun open-log-file ()
  (values NIL "File logging is not supported."))

;;;
;;; DYNAMIC (event) LOGGING DEFINITIONS

(defvar *event-string-buffer* nil
  "The buffer for logging consecutive strings to AppleEvents.
Flush this when we're done and before every PICT.")

(defun open-dynamic-log ()
  (setf *event-string-buffer* NIL)
  (let* ((as-object (choose-and-set-log-application))
         (diag-str (if (null as-object) "Log selection error."
                       "Event logging enabled.")))
    (values as-object diag-str)))

(defun reset-esb ()
  (setf *event-string-buffer* nil))

(defun write-event-string-buffer ()
  (if *event-string-buffer*
    (log-str-or-h-pict *event-string-buffer*))
  (setf *event-string-buffer* NIL))

(defun log-to-event (object)
  #+MCG-TEST
  (progn
    (prin1 object)
    (fresh-line)
    (prin1 (type-of object))
    (fresh-line))
  (typecase object 
    (string (setq *event-string-buffer* (concatenate 'string *event-string-buffer*
                                                     (format nil "~%")
                                                     object)))
    (macptr (setq *event-string-buffer* (concatenate 'string *event-string-buffer*
                                                     (format nil "~%")))
            (write-event-string-buffer)
            (log-str-or-h-pict object))
    (t (log-to-event "bush league log type generated."))))

;; 

(defmethod make-log-item ((object t) &key)
  (format nil "~S" object))

;;
;; Dispatchers for the logging system

(defun prepare-log-destination ()
  "Prepares something to receive logging information based on log preferences."
  (if (eq *log-type* :file)
    (open-log-file)
    (if (eq *log-type* :event)
      (open-dynamic-log)
      (open-test-log))))

(defun write-to-log (item destination)
  "Dispatches a write to the appropriate logging function."
  (case *log-type*
    ((:event) (log-to-event item))  ;; ignore destination
    ((:file) '())    ;; unsupported
    (t (princ item))
    ))

(defun close-log-sink (sink)
  "Cleans up the logging stuff & finalizes the log."
  (if (typep sink 'ASTOOLS:APPLESCRIPT-OBJECT)
    (write-event-string-buffer)))

(defmethod log-list (list)
  "Logs a list of items to the currently specified log destination."
  (if (null list)
    (gui-warn "There is nothing to log.")
    (multiple-value-bind (log-sink message) (prepare-log-destination)
        (if (not log-sink)
          (gui-warn message)
          (progn 
            (loop for x in list do
                  (write-to-log (make-log-item x) log-sink))
            (close-log-sink log-sink))))))


#-ignore
(defmethod log-contents ((notebook clasp-notebook))
  (let ((contents (caddy-contents notebook)))
    (if (null contents)
      (gui-warn "There is nothing in the notebook.")
      (multiple-value-bind (log-sink message) (prepare-log-destination)
        (if (not log-sink)
          (gui-warn message)
          (progn 
            (loop for x in contents do
                  (write-to-log (make-log-item x) log-sink))
            (close-log-sink log-sink)))))))

;;
;; Specialized methods for make-log-item

(defmethod make-log-item ((object comment) &key)
  (comment-contents object))  ;; FIX group comments ?

(defmethod make-log-item ((object clasp-result) &key)
  (clasp-result-text object))

(defmethod make-log-item ((object clasp-command) &key)
  (if (sexp-describe-clasp-command object)
    (format nil "~S" (sexp-describe-clasp-command object))
    (format nil "~S" (sexp-clasp-command object))))

(defmethod make-log-item ((object clasp::variable) &key)
  (format nil "~S" object))

(defmethod make-log-item ((object clasp::dataset) &key)
  (format nil "~S" object))

(defmethod make-log-item ((object string) &key)
  object)

(defmethod make-log-item ((object generic-table) &key (suppress-title nil) (largest-column nil))
  (let ((table-array (array-from-table object))
        (string-table (if suppress-title "" (clasp::name object)))
        (n (if largest-column largest-column (table-largest-datum object))))
    (destructuring-bind (xdim ydim)
                        (array-dimensions table-array)
      (dotimes (y ydim)
        (setq string-table (concatenate 'string string-table
                                        (format nil "~%")))
        (dotimes (x xdim)
          (setq string-table (format nil "~A~A~A"
                                     string-table
                                     (if (null (aref table-array x y))
                                       ""
                                       (substitute #\Space *hdr-char* (string-trim " " (aref table-array x y))))
                                     (if *space-delimit-tables*
                                       (make-string (- n (length (string-trim " " (aref table-array x y))))
                                                    :initial-element #\space)
                                       (string #\tab))
                                     ))))
      (concatenate 'string string-table (format nil "~%")))))

(defmethod make-log-item ((object composite-table) &key)
  (let ((table-string (clasp::name object)))
    (loop for ct in (ct-tables object)
          for ctn in (ct-table-names object)
          do (setq table-string (concatenate 'string
                                             table-string
                                             (format nil "~%~A~%" ctn)
                                             (make-log-item ct :suppress-title t 
                                                            :largest-column (table-largest-datum object)))))
    table-string))

(defmethod make-log-item ((object graph) &key)
  (generate-graph-and-window object :mode :log))

(defun log-selected-items ()
  "interface to the logging facility. if the front-window is the clasp winodw,
it sends the selection to the notebook."
  (let ((fw (front-window))
        (fcw (clasp-window-front)))
    (if (and fcw (equal fw fcw))
      (let ((selected-items (l-history-selected-cw fcw)))
        (setq *clasp-recorded-items* (append *clasp-recorded-items* (reverse selected-items)))))))
