;;;
;;; menu-functions.lisp
;;;

#|
================================================================
Purpose ========================================================
================================================================
Defines utility functions used by menu item actions.


================================================================
Status =========================================================
================================================================
In-progress.


================================================================
Change history =================================================
================================================================
22-May-94 mc	Created.
6-Jun-94 mc	Replaced clasp::variable-value with clasp::simplify .
6-Jul-94  ms    Added compute-scalar-clasp-fcn-on-msel to deal with
                multiple selects in the clasp window.
12-Jul-94 ms    Added optional argument SIMPLIFY to the scalar clasp
                functions to extend their usability to items which should
                not be simplified for the clasp transforms.
7-Sep-94  ms    Most of the compute-scalar-blah-blah are now obsolete, due
                to changes in the way arguments are specified.
|#


(in-package "CLASP-INTERFACE")

(export '(quit-clasp
          select-current-or-new-clasp-window
          edit-preferences))


;;;
;;; Define misc. functions.
;;;

(defun quit-clasp ()
  "Asks the user if she wants to quit and does so, if the answer's yes."
  ;;
  (let ((f-quit-confirmed
         (if *f-confirm-quit*
           (y-or-n-dialog "Really quit CLASP?" :size #@(300 80))
           t)))
    (when f-quit-confirmed
      (quit))))


(defun select-current-or-new-clasp-window ()
  "Selects clasp-window-front if it's non-nil or creates a new one."
  ;;
  (let ((clasp-window-current (clasp-window-front)))
    (if clasp-window-current
      (window-select clasp-window-current)
      (make-instance 'clasp-window))))

(defun gui-warn (string)
  (message-dialog string :size #@(350 150) :ok-text "Ooops"))

(defun warn-feature-missing ()
  "Warns the user that some feature is missing."
  ;;
  (warn "Feature missing"))

;;;
;;; Define preferences-related functions.
;;;

(defun edit-preferences ()
  "Displays a modal dialog in which the user can edit and save CLASP
preferences."
  ;;
  (modal-dialog (make-preferences-dialog) t))

;;;
;;; Define file-related functions.
;;;

(defun choose-and-load-dataset ()
  "Prompts the user for a file then calls clasp:load-dataset on it."
  ;;
  #+DEMO
  (sorry-demo-version)
  #-DEMO
  (let* ((pn-dataset (choose-file-dialog :button-string "Load"))
         (clasp-command (make-clasp-command `(clasp:load-dataset ,pn-dataset))))
    (run-clasp-command clasp-command
                       :str-message (format nil "Loading ~A"
                                            (file-namestring pn-dataset))
                       :f-update-dataset-view t)))


(defun choose-and-import-dataset ()
  "Prompts the user for a file then calls clasp:import-dataset on it."
  ;;
  #+DEMO
  (sorry-demo-version)
  #-DEMO
  (let* ((pn-dataset (choose-file-dialog :button-string "Import"))
         (clasp-command
          (make-clasp-command `(clasp::import-dataset ,pn-dataset
                                :separator ,*char-file-delimiter*
                                :include-labels-p ,*f-include-file-labels*)
                              :standard-error "Unexpected file format. Check import settings in preferences.")))
    (run-clasp-command clasp-command
                       :str-message (format nil "Importing ~A"
                                            (file-namestring pn-dataset))
                       :f-update-dataset-view t)))


(defun export-selected-dataset ()
  "Assumes f-one-dataset-selected-cw returned non-nil on (clasp-window-front) ."
  ;;
  #+DEMO
  (sorry-demo-version)
  #-DEMO
  (let* ((dataset (f-one-dataset-selected-cw (clasp-window-front)))
         (pn-dataset (choose-new-file-dialog :prompt "Export file"
                                             :button-string "Export"))
         (clasp-command
          (make-clasp-command `(clasp::export-dataset ,dataset ,pn-dataset
                                                      :separator ,*char-file-delimiter*
                                                      :include-labels-p ,*f-include-file-labels*))))
    (run-clasp-command clasp-command
                       :str-message (format nil "Exporting ~A"
                                            (file-namestring pn-dataset)))))


(defun save-selected-dataset ()
  "Assumes f-one-dataset-selected-cw returned non-nil on (clasp-window-front)"
  ;;
  #+DEMO
  (sorry-demo-version)
  #-DEMO
  (let* ((dataset (f-one-dataset-selected-cw (clasp-window-front)))
         (pn-dataset (choose-new-file-dialog :prompt "Save file"
                                             :button-string "Save"))
         (clasp-command
          (make-clasp-command `(clasp::save-dataset ',dataset ,pn-dataset))))
    ;;
    (run-clasp-command clasp-command
                       :str-message (format nil "Saving ~A"
                                            (file-namestring pn-dataset)))))


(defun describe-selected-data ()
  "Assumes f-one-data-item-selected-cw returned non-nil on (clasp-window-front)"
  ;;
  (let* ((data-item (f-one-data-item-selected-cw (clasp-window-front)))
         (clasp-command (make-clasp-command `(clasp::description ',data-item))))
    (if data-item
      (run-clasp-command clasp-command
                         :str-message (format nil "Describing ~A" data-item)))))


#|
(defun describe-data (data-item)
  "Describes data to clasp-window-front's status area or via
message-dialog if no clasp-window-front ."
  ;;
  (let ((str-description (clasp::description data-item))
        (clasp-window (clasp-window-front)))
    (cond (clasp-window
           (show-status-view-string clasp-window str-description))
          (t
           (message-dialog str-description :size #@(350 150))))))
|#

#+IGNORE
(defun rename-selected-dataset ()
  "Assumes f-one-dataset-selected-cw returned non-nil on (clasp-window-front)"
  ;;
  (let* ((dataset (f-one-dataset-selected-cw (clasp-window-front)))
         (str-new-name (get-string-from-user
                        (format nil "Enter new dataset name for ~S:" dataset)
                        :initial-string (clasp::name dataset)
                        :ok-text "Rename"))
         (clasp-command
          (make-clasp-command `(clasp::rename-dataset ,dataset ,str-new-name))))
    (clasp::rename-dataset dataset str-new-name)
    ;;
    (run-clasp-command clasp-command
                       :str-message (format nil "Renaming ~S:" dataset)
                       :f-update-dataset-view t)))



(defun rename-selected-things ()
  "Assumes f-one-variable-selected-cw returned non-nil on (clasp-window-front)"
  ;;
  (let ((things (l-data-selected-cw (clasp-window-front))))
    (if (null things)
      (gui-warn "No selected CLASP items to rename.")
      (mapcar #'(lambda (variable)
                  (let* ((str-new-name (get-string-from-user
                                        (format nil "Enter new name for ~S:" variable)
                                        :initial-string (clasp::name variable)
                                        :ok-text "Rename"))
                         (clasp-command
                          (make-clasp-command `(,(if (typep variable 'clasp::variable)
                                                   'clasp::rename-variable
                                                   'clasp::rename-dataset)
                                                ,variable ,str-new-name))))
                    (run-clasp-command clasp-command
                                       :str-message (format nil "Renaming ~S:" variable)
                                       :f-update-dataset-view t)))
              (remove-if-not #'(lambda (thing) (or (typep thing 'clasp::variable) (typep thing 'clasp::dataset))) things)))))


;;;
;;; Define history-related functions.
;;;

(defun log-history-view-selection ()
  "Calls log-item on (clasp-window-front)'s l-history-selected-cw. Assumes
l-history-selected-cw returned non-nil on (clasp-window-front) ."
  ;;
  ;; NOTE: This is commented because
  ;;  it depends on a incremental paste to log application feature.
  ;;  For now we'll use the method of copying the selected
  ;;  history to the clipboard as a PICT, making it easy to paste
  ;;
  #|(loop for item-selected in (l-history-selected-cw (clasp-window-front))
        for f-first-item = t then nil
        do
        (log-item item-selected (if f-first-item t nil)))|#
  (copy-history-as-pict-cw (clasp-window-front)))

;;;
;;; Define open-related functions.
;;;


(defun open-openable-selection ()
  "Calls opens all items in (clasp-window-front)'s
f-selection-openable-cw . Assumes (f-selection-openable-cw (clasp-window-front))
was called and reurned non-nil ."
  ;;
  (let* ((clasp-window (clasp-window-front))
         (l-openable-objects (if clasp-window (l-data-selected-cw clasp-window))))
    (loop for openable-object in l-openable-objects
          do (open-clasp-object openable-object))
    ;; FIX: Add clasp commands?
    ))

;;
;; for the variables menu... X/Y Dock commands
;;  11-29-94 ms no longer forces access to the dock, but sends to frontmost dialog

(defmethod operate-on-dock ((cw WINDOW) (dock-tag SYMBOL))
  (let* ((sourcewin (if (typep cw 'clasp-window) cw
                        (or (find-window "CLASP Dock") NIL)))
         (destwin (if (typep cw 'statistic-dialog-mixin) (prog1 cw (setf sourcewin (or (find-window "CLASP Dock") (clasp-window-front))))
                      (or (and (not (typep sourcewin 'clasp-dock)) (find-window "CLASP Dock"))
                          (find-if #'(lambda (item) (typep item 'clasp-interface::statistic-dialog-mixin)) (windows)))))
         (items (typecase sourcewin
                  (clasp-window (l-data-selected-cw sourcewin))
                  (clasp-dock (items-in-dock dock-tag :select-type :select-or-all))
                  (t nil)))
         )
    (when (and destwin items)
      (typecase destwin
        (clasp-dock (receive-drop (view-named dock-tag destwin) items 0))
        (statistic-dialog-mixin
         (let ((var-pum (ignore-errors (if (eq 'x-var-dock dock-tag) (view-named 'x-var-item destwin)
                                         (if (eq 'y-var-dock dock-tag) (view-named 'y-var-item destwin)
                                             (view-named 'z-var-item destwin))))))
           (when var-pum
             (when (and (not (funcall *cache-key-p*)) (typep var-pum 'variable-pop-up-menu))
               (loop for x in (menu-items var-pum) do
                     (remove-menu-items var-pum x)))
             (receive-drop var-pum items 1))))
        (t NIL)))))

(defmethod operate-on-mrr ((cw WINDOW) (var-tag SYMBOL))
  "operates on the most recent result"
  (let ((var-pum (view-named var-tag cw))
        (selected-items *most-recent-vresult*))
    (if (and var-pum selected-items)
      (progn
        (if (and (not (funcall *cache-key-p*)) (typep var-pum 'variable-pop-up-menu))
          (loop for x in (menu-items var-pum) do
                (remove-menu-items var-pum X)))
        (receive-drop var-pum selected-items 1)
        ))))


(defun dbm ()
  (install-default-menubar)
  (install-clasp-menu))

