;;; new-tables

(in-package :clasp-interface)

;;;
;;; global crap
;;;

(defparameter *hdr-char* #\| "The indicator prefix for table headers (labels).")
(defparameter *ttl-char* #\_ "The indicator prefix for table titles.")
(defvar *clasp-tables* '())
(defparameter *float-control-string* "~,2F")

;;;
;;; table class hierarchy
;;;

;;; mixins

(defclass sort-columns-mixin () ())
(defclass sort-rows-mixin () ())
(defclass sort-table-mixin () ())

(defclass totals-table-mixin () ())

(defclass con-click-mixin ()
  ((c-data :initarg :c-data :accessor c-data)
   (control :initform "Control Table" :initarg :control-title)))
(defclass opt-click-mixin ()
  ((o-data :initarg :o-data :accessor o-data)
   (option :initform "Option Table" :initarg :option-title)))
(defclass dbl-click-mixin ()
  ((d-data :initarg :d-data :accessor d-data)
   (double :initform "Double Table" :initarg :double-title)))

(defclass graphable-table-superclass () ())

(defclass plot-rows-mixin (graphable-table-superclass) ())
(defclass plot-columns-mixin (graphable-table-superclass) ())

;;; actual hierarchy

(defclass clasp-table (clasp::named-object-mixin) 
  ((prepared-table :initform NIL  :initarg :prepared-table :accessor prepared-table)
   (largest-datum :initform NIL :type FIXNUM :accessor table-largest-datum)))

(defclass composite-table (clasp-table) 
  ((tables :initform NIL :initarg :tables :accessor ct-tables)
   (table-names :initform NIL :initarg :table-names :accessor ct-table-names)))

(defclass generic-table (clasp-table)
  ((column-names :initform nil :initarg :column-names :type LIST :accessor table-column-names)
   (row-names :initform nil :initarg :row-names :type LIST :accessor table-row-names)
   (table-data :initform nil :initarg :table-data :type LIST :accessor table-data)
   #+NOW-INHERITED (prepared-table :initform nil :type LIST :accessor prepared-table)
   #+NOW-INHERITED (largest-datum :initform NIL :type FIXNUM :accessor table-largest-datum)))

(defclass group-means-table (generic-table sort-columns-mixin sort-rows-mixin plot-rows-mixin) ())

(defclass group-means-confi-table (group-means-table) 
  ((confidence-interval-groups :initform NIL :initarg :confi-groups :accessor confidence-interval-groups)
   (group-names :initform NIL :initarg :group-names :accessor confidence-group-names)))

(defclass scheffe-table (generic-table con-click-mixin opt-click-mixin) ())

(defclass anova-table (generic-table) ())

(defclass contingency-table (generic-table totals-table-mixin sort-columns-mixin sort-rows-mixin con-click-mixin opt-click-mixin dbl-click-mixin)
  ()
  (:default-initargs :double-title "Observed Frequencies"
    :option-title "Column Margins"
    :control-title "Row Margins"))

(defclass lag-table (generic-table plot-rows-mixin) ())

(defclass correlation-table (generic-table) ())

(defclass t-table (generic-table) ())
(defclass d-table (generic-table) ())
(defclass chi-square-table (generic-table) ())

;;; correlation-table, lag-table

;;;
;;; generics
;;;

(defgeneric fill-out-table (table))
(defgeneric prep-table (table))
(defgeneric sort-table-columns (table))
(defgeneric sort-table-rows (table))
(defgeneric calc-totals (table))
(defgeneric calc-control-table (table))
(defgeneric calc-option-table (table))
(defgeneric calc-double-table (table))

;;;
;;; print object
;;;

(defmethod print-object ((table generic-table) stream)
  (format stream "#.(TABLE '~S ~S)" (type-of table) (clasp::name table)))

;;;
;;; princ-to-table-string
;;;

(defun table-princ (datum stream &optional (pad-to NIL))
  (if datum
    (let ((raw (format nil
                       (typecase datum
                         (float *float-control-string*)
                         (ratio *float-control-string*)
                         (t "~A"))
                       datum)))
      (format stream "~A~A"
              (if (and pad-to (> pad-to (length raw)))
                (make-string (- pad-to (length raw)) :initial-element #\space) "")
              raw))
    NIL))

;;;
;;; largest-column
;;;

(defmethod largest-datum (list)
  (labels ((flatten (list)
             (if (null list) NIL
                 (if (atom (car list))
                   (cons (car list) (flatten (cdr list)))
                   (append (flatten (car list)) (flatten (cdr list)))))))
    (loop for i in (flatten list) maximize (length (table-princ i nil)))))

;;;
;;; initialize methods
;;;

(defmethod initialize-instance :around ((table generic-table) &key (primary-table NIL))
  (call-next-method)
  (calc-totals table)
  (calc-double-table table)
  (calc-option-table table)
  (calc-control-table table)
  (prep-table table)
  (when (and primary-table *AUTO-OPEN-PRIMARY-TABLES*)
    (open-clasp-object table)))

(defmethod initialize-instance :after ((table generic-table) &key)
  (fill-out-table table))

(defmethod initialize-instance :after ((table sort-rows-mixin) &key)
  (sort-table-rows table))

(defmethod initialize-instance :after ((table sort-columns-mixin) &key)
  (sort-table-columns table))

(defmethod initialize-instance :after ((table composite-table) &key (primary-table t))
  (prep-table table)
  (when (and primary-table *auto-open-primary-tables*)
    (open-clasp-object table)))

;;;
;;; fill-out-table : fill in missing cells
;;;

(defmethod fill-out-table ((table generic-table))
  (let ((n-cols (max (length (table-column-names table)) (loop for i in (table-data table) maximize (length i))))
        (n-rows (max (length (table-row-names table)) (length (table-data table)))))
    (labels ((fill-rows (rows)
               (if (null rows)
                 NIL
                 (cons (append (car rows) (make-list (- n-cols (length (car rows))) :initial-element NIL))
                       (fill-rows (cdr rows))))))
      (setf (table-data table) 
            (append (fill-rows (table-data table))
                    (make-list (- n-rows (length (table-data table)))
                               :initial-element (make-list n-cols :initial-element NIL)))))))

;;;
;;; sort-table methods
;;;

(defun robust-lt (e1 e2)
  (flet ((terrorize (v)
           (typecase v
             (number v)
             (string v)
             (t (princ-to-string v)))))
    (let* ((me1 (terrorize e1))
           (me2 (terrorize e2)))
      (if (and (numberp me1) (numberp me2))
        (< me1 me2)
        (string< (princ-to-string me1) (princ-to-string me2))))))

(defmethod sort-table-rows ((table generic-table))
  (let ((comb-table (if (table-row-names table)
                      (mapcar #'cons (table-row-names table) (table-data table))
                      (return-from sort-table-rows nil))))
    (setf comb-table (sort comb-table #'robust-lt :key #'car))
    (setf (table-data table) (mapcar #'cdr comb-table))
    (setf (table-row-names table) (mapcar #'car comb-table))))

(defmethod sort-table-columns ((table generic-table))
  (let ((comb-table (if (table-column-names table)
                      (mapcar #'cons (table-column-names table) (clasp::transpose (table-data table)))
                      (return-from sort-table-columns nil))))
    (setf comb-table (sort comb-table #'robust-lt :key #'car))
    (setf (table-data table) (clasp::transpose (mapcar #'cdr comb-table)))
    (setf (table-column-names table) (mapcar #'car comb-table))))

;;;
;;; calculate totals
;;;

(defmethod calc-totals ((table t)) NIL)

(defmethod calc-totals ((table totals-table-mixin))
  (let ((ctotals (loop for i from 0 to (1- (length (car (table-data table)))) collect
                       (loop for j in (table-data table) sum
                             (or (ignore-errors (elt j i)) 0))))
        (rtotals (loop for i in (table-data table) collect
                       (reduce #'+ i))))
    (setf (table-data table)
          (append (mapcar #'(lambda (r rt) (append r (cons rt nil))) (table-data table) rtotals)
                  (list (append ctotals (list (reduce #'+ ctotals))))))
    (setf (table-row-names table) (append (table-row-names table) (list "Total")))
    (setf (table-column-names table) (append (table-column-names table) (list "Total")))
    table))

;;;
;;; caclucate secondary tables
;;;

(defmethod calc-double-table ((table t)) NIL)
(defmethod calc-control-table ((table t)) NIL)
(defmethod calc-option-table ((table t)) NIL)

(defmethod calc-double-table ((table dbl-click-mixin))
  (setf (d-data table) (table-data table)))

(defmethod calc-control-table ((table contingency-table))
  (flet ((calc-row-percents (data)
           (mapcar #'(lambda (list)
                       (let ((maxi (length list)) lbuff)
                         (dotimes (i (1- maxi))
                           (setf lbuff (append lbuff (list (/ (elt list i) (elt list (1- maxi)))))))
                         (setf lbuff (append lbuff (list 1.0)))
                         lbuff))
                   data)))
    (setf (c-data table) (calc-row-percents (table-data table)))))

(defmethod calc-option-table ((table contingency-table))
  #+OLD (flet ((calc-col-percents (data)
                 (let (buff result)
                   (dotimes (i (1- (length data)))
                     (setf buff '())
                     (dotimes (ii (length (elt data i)))
                       (setf buff (append buff (list (/ (elt (elt data i) ii)
                                                        (elt (car (last data)) ii))))))
                     (setf result (append result (list buff))))
                   (append result (list (make-list (length (car data)) :initial-element 1.0)))))))
  (flet ((calc-col-percents (data)
           (append 
            (loop for row in data collect
                  (let ((total (car (last row))))
                    (loop for v in row collect
                          (/ v total))))
            (list (make-list (length (car data)) :initial-element 1)))))
    (setf (o-data table) (calc-col-percents (table-data table)))))

;;;
;;; prepare table for display
;;;

(defmethod prep-table ((table generic-table))
  (setf (table-largest-datum table)
        (largest-datum (list (table-row-names table)
                             (table-column-names table)
                             (table-data table))))
  (let ((row-prepared-table 
         (if (table-row-names table)
           (mapcar #'cons (mapcar #'(lambda (i) (format nil "~C~A" *hdr-char* i)) (table-row-names table))
                   (table-data table))
           (table-data table))))
    (setf (prepared-table table)
          (if (table-column-names table)
            (if (table-row-names table) 
              (cons (cons NIL (mapcar #'(lambda (i) (format nil "~C~A" *hdr-char* i)) (table-column-names table))) row-prepared-table)
              (cons (mapcar #'(lambda (i) (format nil "~C~A" *hdr-char* i)) (table-column-names table)) row-prepared-table))
            row-prepared-table))
    (setf (prepared-table table)
          (mapcar #'(lambda (row) (mapcar #'(lambda (item) (table-princ item nil (table-largest-datum table))) row))
                  (prepared-table table)))))

(defmethod prep-table ((table composite-table))
  (setf (prepared-table table) 
        (loop for tab in (ct-tables table)
              for tabn in (ct-table-names table) append
              (progn
                ;; this is UGLY ! it side effects the original table (blech)
                (prep-table tab)
                (append (list NIL (list (format nil "~C~A" *ttl-char* tabn))) (prepared-table tab)))))
  (let ((maxrowlen (loop for r in (prepared-table table) maximize (length r))))
    (setf (prepared-table table)
          (mapcar #'(lambda (row)
                      (if (< (length row) maxrowlen)
                        (append row (make-list (- maxrowlen (length row)) :initial-element NIL))
                        row))
                  (prepared-table table))))
  (setf (table-largest-datum table)
        (largest-datum (prepared-table table))))

;;;
;;; click table
;;;

(defmethod control-click-table ((window t) (table generic-table))
  NIL)
(defmethod option-click-table ((window t) (table generic-table))
  NIL)
(defmethod double-click-table ((window t) (table generic-table))
  NIL)
(defmethod control-click-table ((window t) (table con-click-mixin))
  (setf (table-data table) (c-data table))
  (prep-table table)
  (revalidate-view window))
(defmethod option-click-table ((window t) (table opt-click-mixin))
  (setf (table-data table) (o-data table))
  (prep-table table)
  (revalidate-view window))
(defmethod double-click-table ((window t) (table dbl-click-mixin))
  (setf (table-data table) (d-data table))
  (prep-table table)
  (revalidate-view window))

;;;
;;; ACTUAL INTERFACE STUFF
;;;

;;; VIEW

(defclass table-view (array-dialog-item)
  ((statistical-table
    :initform nil
    :initarg :table
    :type STATISTICAL-TABLE
    :accessor table-view-table
    :documentation "the statistical table being viewed.")
   (view-array
    :initform nil
    ;; init'd by intialize instance
    :type LIST
    :accessor view-array-table-view
    :documentation "the actual viewing stuff."))
    (:documentation "an array-dialog-item subclass for showing a table's contents.")
    (:DEFAULT-INITARGS
      :view-font *table-header-font*
      :table-hscrollp t
      :table-vscrollp t
      :table-print-function
      #'(lambda (print-item stream)
          (declare (ignore stream))
          (let ((font-spec (cond
                            ((and print-item (find *hdr-char* print-item)) *table-header-font*)
                            ((and print-item (find *ttl-char* print-item)) *table-title-font*)
                            (t *table-entry-font*))))
            (multiple-value-bind (ff ms) (font-codes font-spec)
              (with-font-codes ff ms
                (when print-item
                  (with-pstrs ((p-str (remove *ttl-char* (remove *hdr-char* (table-princ print-item nil)))))
                    (#_DrawString p-str)))))))))

(defmethod initialize-instance :after ((table-view table-view) &key)
  "sets the cell size to font-sensitive width, then sets up the array of numbers for the array subview"
  (set-cell-size table-view 
                 (string-width (make-string (table-largest-datum (table-view-table table-view)) :initial-element #\9) 
                               *table-header-font*)
                 (table-char-height))
  (set-table-array table-view
                   (array-from-table (table-view-table table-view))))

(defmethod revalidate-view ((table-view table-view))
  (set-cell-size table-view 
                 (string-width (make-string (table-largest-datum (table-view-table table-view)) :initial-element #\9) 
                               *table-header-font*)
                 (table-char-height))
  (set-table-array table-view
                   (array-from-table (table-view-table table-view))))

(defmethod array-from-table ((table clasp-table))
  (let* ((cols (loop for i in (prepared-table table) maximize (length i)))
         (rows (length (prepared-table table)))
         (array (make-array (list cols rows) :initial-contents (clasp::transpose (prepared-table table)))))
    array))

;;; WINDOW

(defclass table-view-window (window)
  ()
  (:documentation "A window subclass used for viewing statistical tables."))

(defmethod window-size-parts :after ((table-view-window table-view-window))
  "Sets the array-dialog-item to fill the window."
  (let ((table-array-item (view-named 'table-array-item table-view-window)))
    (when table-array-item
      (set-view-position table-array-item 0 0)
      (set-view-size table-array-item (view-size table-view-window)))))

(defmethod initialize-instance :after ((table-view-window table-view-window)
                                           &key table)
  "Adds the actual table view to the window."
  (add-subviews table-view-window
                (make-instance 'table-view :view-nick-name 'table-array-item
                               :table table))
  (window-size-parts table-view-window))

(defmethod view-activate-event-handler :after ((window table-view-window))
  (ribbon-display (window-title window)))

;;; OPEN OBJECT

(defun table-char-height ()
  (+ (car (last *table-entry-font*)) 2))
(defun table-char-width ()
  (car (last *table-entry-font*)))

(defmethod f-can-open-clasp-object ((table clasp-table))
  t)

(defmethod open-clasp-object ((table clasp-table))
  (let ((win (find-if #'(lambda (window)
                          (ignore-errors (equal table (table-view-table (view-named 'table-array-item window)))))
                      (windows))))
    (if win
      (window-select win)
      (let* ((maxcollen (table-largest-datum table))
             (cols (loop for r in (prepared-table table) maximize (length r)))
             (rows (length (prepared-table table)))
             ;; screen x width and y width. note font-sensitive x spacing.
             (sx (min *screen-width* (+ 16 (* (string-width (make-string maxcollen :initial-element #\9) *table-header-font*)
                                              cols))))
             (sy (min *screen-height* (+ 30 (* (table-char-height) (1+ rows))))))
        (make-instance 'table-view-window 
          :window-title (trim-to-maxlen (CLASP::name table) *window-title-maximum-length*)
          :table table
          :view-size
          (make-point sx sy)
          :view-position (best-spot sx sy))))))

;;;
;;; plotting stuff
;;;

;; generic-do-nothing cases for basic tables
(defmethod x-labels ((table clasp-table))
  '()) ;; return nil to indicate that we don't care about labels
(defmethod generate-plots ((table clasp-table))
  '())

;;; functions that actually do something

;;; contingency tables
(defmethod generate-plots ((table contingency-table))
  (let* ((data (mapcar #'(lambda (item) (subseq item 0 (1- (length item))))
                       (subseq (table-data table) 0 (- (length (table-data table)) 1))))
         (rlen (length data))
         used-colors buff plots)
    (dotimes (x rlen)
      (push (cons (list :polyline-with-vertices (format nil "Frequency where ~:(~A~) = ~A"
                                                        (clasp::name (slot-value table 'y-variable))
                                                        (remove (aref *hdr-char* 0) (elt (table-row-names table) x)))
                        (list :color (setf buff (new-color used-colors))
			      :symbol (svref *symbologies* x)))
                  (mapcar #'list
                          (make-count-list 0 (1- (length (elt data x)))) (elt data x)))
            plots)
      (push buff used-colors))
    plots))

(defmethod x-labels ((table contingency-table))
  (let ((labels (table-column-names table))
        l-a-list)
    (dotimes (i (1- (length labels)))
      (push (list (remove (aref *hdr-char* 0) (elt labels i)) i)
            l-a-list))
    (reverse l-a-list)))

;;; group means tables
; 2-2-96 was group-means - Westy
(defmethod generate-plots ((table group-means-table))
  (let* ((data (if (search "Total" (car (last (table-column-names table))))
                 (mapcar #'(lambda (item) (subseq item 0 (1- (length item))))
                         (subseq (table-data table) 0 (- (length (table-data table)) 1)))
                 (table-data table)))
         (rlen (length data))
         (ci (confidence-intervals table))
         used-colors buff plots)
    (dotimes (x rlen)
      (push (cons (list (if ci :polyline-with-confidence-intervals :polyline-with-vertices) 
                        (concatenate 'string 
                                     (string-left-trim "|" (nth x (table-row-names table)))
                                     " " 
                                     (string-capitalize (clasp::name table)))
                        (append (list :color (setf buff (new-color used-colors)))
                                `(:symbol ,(svref *symbologies* x))
                                (if ci `(:vertices? t) '())))
                  (if ci
                    (mapcar #'cons
                            (make-count-list 0 (1- (length (elt data x)))) ci)
                    (mapcar #'list
                            (make-count-list 0 (1- (length (elt data x)))) (elt data x))))
            plots)
      (push buff used-colors))
    plots))

; 2-2-96 was group-means - Westy
(defmethod x-labels ((table group-means-table))
  (let ((labels (table-column-names table))
        l-a-list)
    (dotimes (i (length labels))
      (unless (search "Total" (elt labels i))
        (push (list (remove (aref *hdr-char* 0) (or (ignore-errors (string-trim "= " (subseq (elt labels i) (position #\= (elt labels i)) (length (elt labels i)))))
                                                    (elt labels i))) i)
            l-a-list)))
    (reverse l-a-list)))

;;; lagged tables
(defmethod generate-plots ((table lag-table))
  (let ((data (mapcar #'car (table-data table))))
    (list (cons (list :polyline-with-vertices (format nil "~:(~A~)" (clasp::name table))
                      (list :color (new-color '())))
                (mapcar #'(lambda (item1 item2)
                            (list (parse-float (subseq item2 (1+ (search "g" item2))))
                                  item1))
                        data (table-row-names table))))))

