;;;; **************************************************************************
;;;; **************************************************************************
;;;; *                                                                        *
;;;; *                             DIALOG PACKAGE                             *
;;;; *                                                                        *
;;;; **************************************************************************
;;;; **************************************************************************
;;;
;;; Written by: Matt Schmill
;;;             Department of Computer and Information Science
;;;             University of Massachusetts
;;;             Amherst, Massachusetts 01003.
;;;
;;;  04-12-95 File Created.  (Schmill)
;;;

;;; TO DO: destroy the frames on exit
;;;        make the ok-function work on modals

;;; package stuff

(defpackage :dialog
  (:use clim clim-lisp))
(in-package :dialog)

;; exports

(export '(dialog
	  get-dialog-gadget-values get-gadget get-gadget-value get-gadget-items
	  add-to-list-pane reset-list-pane
	  make-dialog))

;;; size parameters
(defparameter *cancel-button-width* 70)
(defparameter *button-height* 35)

;;; the dialog frame type
(define-application-frame dialog ()
  ((gadget-alist :initform nil :accessor dialog-gadget-alist)
   (output-record :initform NIL :accessor dialog-output-record)
   (last-message :initform "" :accessor dialog-last-message))
  (:panes
   (main :application :scroll-bars NIL :end-of-line-action :allow))
  (:layouts
   (default (vertically () main))))

(defmethod frame-standard-output ((dialog dialog))
  (get-frame-pane dialog 'main))

;;; a text viewer frame type
(define-application-frame text-viewer ()
  ((text :initform nil :initarg :text :accessor tv-text)
   (kill-me :initform nil :accessor tv-kill-me))
  (:panes (viewer :application :scroll-bars :vertical :display-function 'tv-refresh
                  :incremental-redisplay t :min-width 300 :min-height 180
                  :width 320 :height 200)
   (ok-button push-button :label "OK" :activate-callback
              #'(lambda (b) (setf (tv-kill-me (pane-frame b)) t)
                  (frame-exit (pane-frame b)))))
  (:layouts (def (vertically () viewer ok-button))))

(defmethod tv-refresh ((self text-viewer) stream)
  (filling-output (stream :fill-width (list (window-inside-width stream) :pixel))
    (princ (tv-text self) stream)))

(defun text-view (text)
  (let ((tv (make-application-frame 'text-viewer
	      :width 400 :height 300
              :text text)))
    (run-frame-top-level tv)))

;;
;; functions for getting the gadget values
;;

(defmethod get-dialog-gadget-values ((dialog dialog))
  "returns a list of the dialog's gadget values. ignores
gadgets which do not inherit from clim::value-gadget."
  (nreverse (delete :no-value 
                    (mapcar #'(lambda (g) (if (typep (cdr g) 'clim::value-gadget)
			                      (gadget-value (cdr g))
                                            :no-value))
                            (dialog-gadget-alist dialog)))))

(defmethod get-gadget-value ((dialog dialog) gad-id)
  "Gets a gadget value by gadget-tag."
  (let ((gad (cdr (assoc gad-id (dialog-gadget-alist dialog)))))
    (if (typep gad 'clim::value-gadget)
        (gadget-value gad)
      :no-value)))

(defmethod get-gadget-items ((dialog dialog) gad-id)
  "Gets the items in a list pane type dialog."
  (let ((gad (cdr (assoc gad-id (dialog-gadget-alist dialog)))))
    (ignore-errors (set-gadget-items gad))))

;;
;; function for getting at a gadget pane

(defmethod get-gadget ((dialog dialog) gad-id)
  "Returns a gadget pane. Not encouraged for use."
  (cdr (assoc gad-id (dialog-gadget-alist dialog))))

;;
;; macro functions for defining things like radio boxes
;;

;; lispworks sucks at radio boxes 4-12-95 ms
#+CRASHES
(defun construct-radio-cluster (stream titles)
  "takes a list of toggle button titles and makes a radio box pane."
  (warn "Lispworks has trouble with radio boxes !")
  (let ((panes-in-the-ass (mapcar #'(lambda (name) 
                                      (with-output-as-gadget
                                          (stream)
                                        (make-pane 'toggle-button
						   :label name)))
                                  titles)))
    (make-pane 'radio-box :choices panes-in-the-ass)))

;;
;; some support functions for operating on abstract gadgets

(defmethod append-to-list-pane ((dialog dialog) pane-symbol new-items &key (position :first)
                                (refresh t))
  "Adds a new item to a list pane."
  (let ((pane (cdr (assoc pane-symbol (dialog-gadget-alist dialog)))))
    (case position
      (:first (setf (clim:set-gadget-items pane)
                    (append new-items (clim:set-gadget-items pane))))
      (:last (setf (clim:set-gadget-items pane) 
                   (append (clim:set-gadget-items pane) new-items))))
    ;; refresh
    (when refresh
      (repaint-sheet pane (sheet-region pane))) ;; call repaint sheet, i guess
    ))

(defmethod add-to-list-pane ((dialog dialog) pane-symbol new-item &key (position :first)
                             (refresh t))
  "Adds a new item to a list pane."
  (let ((pane (cdr (assoc pane-symbol (dialog-gadget-alist dialog)))))
    (case position
      (:first (push new-item (clim:set-gadget-items pane)) (princ "ASAS"))
      (:last (setf (clim:set-gadget-items pane) 
                   (append (clim:set-gadget-items pane) (list new-item)))))
    ;; refresh
    (when refresh
      (repaint-sheet pane (sheet-region pane))) ;; call repaint sheet, i guess
    ))

(defmethod reset-list-pane ((dialog dialog) pane-symbol new-item-list &key (refresh t))
  (let ((pane (cdr (assoc pane-symbol (dialog-gadget-alist dialog)))))
    (setf (clim:set-gadget-items pane) new-item-list)
    ;; refresh
    (when refresh
      (repaint-sheet pane (sheet-region pane)))))

(defun always-ok (&rest args)
  (declare (ignore args))
  t)

(defmethod message ((dialog dialog) message)
  (let ((stream (frame-standard-output dialog)))
    (with-text-face (stream :bold)
      (with-drawing-options (stream :ink +flipping-ink+)
        (stream-set-cursor-position stream *cancel-button-width*
                                    (- (window-inside-height stream) 20))
        (format (frame-standard-output dialog) (dialog-last-message dialog)))
      (setf (dialog-last-message dialog) message)
      (stream-set-cursor-position stream *cancel-button-width*
				  (- (window-inside-height stream) 20))
      (format (frame-standard-output dialog) (dialog-last-message dialog)))))

(defmethod warn-not-ok ((dialog dialog) why-string)
  (beep)
  (message dialog why-string))
  
;;
;; THE BIG SUCKER

(defparameter *cf* '())

(defun make-dialog (&key (label NIL) (frame-args NIL) (gadget-defs NIL) 
                         (parent NIL) (decorate NIL) (pad-x 0) (pad-y 0)
                         (ok-function #'null-action) (kill-on-ok nil) (destroy t)
                         (modal NIL) (ok-filter #'always-ok))
  "Actually makes, displays and maintains the dialog. See the docs for details."
  (let* ((waiting-for-ok modal) (cancel-flag NIL)
         (d-top (if label 20 0)) 
         ;; space requirements
         (d-width (+ (reduce #'(lambda (a b) (if (< a b) b a))
			     (mapcar #'(lambda (gd)
				         (+ (or (getf gd :left) 0)
					    (if (integerp (getf gd :width))
					        (getf gd :width) 50)))
				     gadget-defs)) 5 pad-x))
	 (d-height (+ (reduce #'(lambda (a b) (if (< a b) b a))
			      (mapcar #'(lambda (gd) 
					  (+ (or (getf gd :top) 0)
					     (if (integerp (getf gd :height))
						 (getf gd :height) 30)))
				      gadget-defs)) *button-height* d-top 20 pad-y))
         ;; tweak the frame options
	 (revised-opts (append (when (not (find :width frame-args)) (list :width d-width))
			       (when (not (find :height frame-args)) 
                                 (list :height d-height))
			       frame-args))
         (port (find-port))
         (framem (typecase port
                   (frame-manager port)
                   (t (find-frame-manager :port port))))
	 (new-frame (apply #'make-application-frame 'dialog
                           :frame-manager framem
			   revised-opts))
         (stream (get-frame-pane new-frame 'main))
	 buff temp-dg gd)
    (setq *cf* new-frame)
    ;; open the window
    (mp::process-run-function (format nil "Dialog frame")
			      '(:run-reasons :enable)
			      #'run-frame-top-level new-frame)
    ;; draw decorations
    (when decorate
      (funcall decorate stream))
    ;; draw the label
    (when label 
      (with-text-face (stream :bold)
	(format stream label)))
    ;; make the explicit gadgets into dialog-gadgets
    (loop for gd1 in gadget-defs do
	  ;; make a copy of the gadget def so we don't destroy things
	  (setq gd (copy-list gd1))
	  ;; set stream cursor
	  (stream-set-cursor-position stream (or (getf gd :left) 0)
				      (+ d-top (or (getf gd :top) 0)))
	  ;; catch the :width :fill option
	  (when (eql (getf gd :width) :fill)
	    (setf (getf gd :width) (- (window-inside-width stream) (or (getf gd :left) 0))))
	  ;; remove the extra supported options
	  (remf gd :left) (remf gd :top)
	  ;; make the pane and draw it immediately
	  (with-output-as-gadget (stream)
	    (setf temp-dg (apply #'make-pane (cdr gd))))
	  ;; push the new dg onto the gad database
	  (push (cons (first gd) temp-dg) (dialog-gadget-alist new-frame)))
    ;; make the implicit push-buttons
    (stream-set-cursor-position stream 0 (- (window-inside-height stream) (- *button-height* 4)))
    (with-output-as-gadget (stream)
      (setq buff (make-pane 'push-button :label "OK"
                            ;; the ac's check the ok-filter, then act accordingly
			    :activate-callback (if (not modal)
						   #'(lambda (button)
                                                       (multiple-value-bind (ok why)
                                                           (funcall ok-filter (pane-frame button))
						         (if ok
                                                             (progn
						               (funcall ok-function new-frame)
                                                               (when kill-on-ok
						                 (frame-exit (pane-frame button))))
                                                           (warn-not-ok (pane-frame button) why))))
						 #'(lambda (button)
                                                       (multiple-value-bind (ok why)
                                                           (funcall ok-filter (pane-frame button))
						         (if ok
                                                             (progn
						               (setf waiting-for-ok nil)
						               (frame-exit (pane-frame button)))
                                                           (warn-not-ok (pane-frame button) why))))))))
    (push (cons 'the-ok-button buff) (dialog-gadget-alist new-frame))
    (stream-set-cursor-position stream (- (window-inside-width stream) *cancel-button-width*)
				(- (window-inside-height stream) (- *button-height* 4)))
    (with-output-as-gadget (stream)
      (setq buff (make-pane 'push-button :label "CANCEL"
			    :activate-callback #'(lambda (button)
						   (setf waiting-for-ok nil)
						   (setf cancel-flag t)
						   (frame-exit (pane-frame button))))))
    (push (cons 'the-cancel-button buff) (dialog-gadget-alist new-frame))
    ;; busy poll or exit
    (when modal (loop while waiting-for-ok do (sleep 0.3)))
    (if (not modal)
        new-frame
      (if cancel-flag
          (values :CANCEL)
	(get-dialog-gadget-values new-frame)))
    ))

;;
;; a stub
;;

(defun null-action (&rest args)
  ())

