;;;; -*- Mode:Common-Lisp; Package:EKSL-UTILITIES; Fonts:(medfnt courierfb hl12i tr12 medfnb cptfonti hl12b); Base:10 -*-
;;;; *-* File: /usr/users/eksl/systems/utils/dev/priority-queue.lisp *-*
;;;; *-* Last-edit: Wednesday, July 5, 1995  17:12:55; Edited-By: anderson *-* 
;;;; *-* Machine: Christa (Explorer II, Microcode 489) *-*
;;;; *-* Software: TI Common Lisp System 6.49 *-*
;;;; *-* Lisp: TI Common Lisp System 6.49  *-*

;;;; **************************************************************************
;;;; **************************************************************************
;;;; *                                                                        *
;;;; *                         Priority Queue Routines                        *
;;;; *                                                                        *
;;;; **************************************************************************
;;;; **************************************************************************
;;;
;;; Written by: David L. Westbrook and Mike Greenberg
;;;             Department of Computer and Information Science
;;;             University of Massachusetts
;;;             Amherst, Massachusetts 01003.
;;;
;;; * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
;;;
;;;  09-20-89 Added `priority-queue-empty', `priority-queue-empty-p'
;;;           and `priority-queue-map' functions.  (Westy)
;;;  12-07-89 Fixed (defstruct (priority-queue... to use :PRINT-FUNCTION
;;;           instead of :PRINT.  (Westy)
;;;  02-22-91 Reimplemented to allow priority-queue resources and and more
;;;           initialization options.  (Westy)
;;;  04-08-92 Removed use of `temp-vector' resource for the heap.  (WESTY)
;;;  03-19-93 Added `priority-queue-find.'  We should make these things work 
;;;           like sequences with extra abilities.  (Anderson)
;;;  12-22-93 Added a non-Explorer definition for `with-priority-queue-locked.'
;;;           This code should be portable.  (anderson)
;;;  12-22-93 Changed a use of `utils:for' into a loop, for portability.
;;;  12-22-93 Changed `lsh' to `ash,' since `ash' is Common Lisp.  (anderson)
;;;           (anderson)
;;;  03-17-94 Removed the Explorer-specific resource stuff.  (anderson)
;;;  03-17-94 Added an EXPORT at the top of the file.  (anderson)
;;;  06-09-94 Removed slot documentation, since that's not Common Lisp.  (Anderson)
;;;  06-28-95 Exported `reorder-priority-queue.'  (anderson)
;;;
;;; * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
;;; --*--

(in-package :eksl-utilities)

(export '(make-priority-queue
	   priority-queue-insert
	   priority-queue-pop
	   priority-queue-head
	   priority-queue-head-or-nil
	   priority-queue-length
	   priority-queue-delete
	   priority-queue-find
	   priority-queue-empty-p
	   priority-queue-empty
	   priority-queue-map-in-priority-order
	   priority-queue-map-in-heap-order
	   priority-queue-elements
	   print-priority-queue-elements
	   do-priority-queue
	   reorder-priority-queue))

(proclaim '(optimize (speed 3) (safety 1)))

;;; --*--
;;; ***************************************************************************
;;;
;;; Define a priority queue datatype
;;;  A priority queue is represented as a structure.
;;;  The queue is implemented as a variable length heap.
;;;  Insert & remove time are log N.
;;;  This is quicker than simple list implementations for
;;;  17 < n, for totally random inserts
;;;  When inserts are toward the end of the queue, the tradeoff is at about 3
;;;
;;;  This also does no consing.
;;;
;;;  Functions are
;;;  (make-priority-queue not-lessp-predicate
;;;                       &key key initial-contents (size 100)
;;;                            element-type disable-locking 
;;;                            process-queue-length)
;;;  (priority-queue-insert element queue)
;;;  (priority-queue-pop  queue)
;;;  (priority-queue-head queue)
;;;  (priority-queue-head-or-nil queue)
;;;  (priority-queue-length queue)
;;;  (priority-queue-empty-p queue)
;;;  (priority-queue-empty queue)
;;;  (priority-queue-map-in-priority-order function priority-queue &rest extra-args)
;;;  (priority-queue-map-in-heap-order queue)
;;;  (priority-queue-elements queue)
;;;  (priority-queue-find item priority-queue &key key test)
;;;  (priority-queue-delete item priority-queue)
;;;  (clear-priority-queue queue)
;;;  (do-priority-queue (var queue) &body body)
;;;  (print-priority-queue queue stream depth)
;;;  (print-priority-queue-elements queue &optional stream)
;;;  (reorder-priority-queue queue)
;;;
;;;----------------------------------------------------------------------------
;;;
;;; Important note: The priority queue heap array is one based.

(defstruct (priority-queue
	    (:constructor make-priority-queue-internal)
	    (:print-function print-priority-queue))
  heap
  (length 0)
  (key-function  nil)
  (sort-function nil)
  ;; T (semaphore), :QUEUED or NIL (disabled)
  (locking-style t)
  (lock nil)
  ;; In case we want to add another lock and don't want to recompile everything.
  #+ignore
  (unused nil :documentation "Currently unused"))

(setf (documentation 'priority-queue 'structure)
      "Implements a heap-based priority queue.  Do an apropos on ``priority-queue'' for
a list of functions to operate on this data structure.")

(defun make-priority-queue (not-lessp-predicate &key
                               key initial-contents (size 100) (element-type t)
                               disable-locking process-queue-length)
  "Creates and returns a new priority queue.  Elements of the queue are ordered
by the dyadic `not-lessp-predicate,' which is true iff its first argument is of
equal or higher priority than its second argument.  \(EFFICIENCY NOTE: if the
two arguments are of equal priority, it is semantically equivalent, but
implementationally SLOWER to return nil.  For example, use <=, not <.\) `key' is
a monadic function which maps an element of the queue into an argument for
`not-lessp-predicate.' This is analogous to the :key argument of generic
sequence functions such as `sort,' and similarly defaults to the identity
function.  The queue is defined so that the element returned by
`priority-queue-pop' is the \(an\) element X such that for all Y in the queue,
\(not-lessp-predicate \(key X\) \(key Y\)\).  If supplied, `initial-contents'
should be a sequence of elements to insert into the queue.  The queue can be
initialized this way in O\(n\) time, while N insertion operations will take O\(n
log n\) time.  `size' determines the initial size of the queue.  `element-type'
can be used to specify the allowed type of elements.  `disable-locking' if
non-nil will provide reasonable time savings.  Use when exclusive use is not a
problem.  `process-queue-length' if non-nil and `disable-locking' is nil can be
used to specify the size of the process queue used during locking.  If this is
nil or zero a simple semaphore method will be used."
  (values
    (initialize-priority-queue
      (make-priority-queue-internal)
      nil
      initial-contents
      element-type
      size
      key
      not-lessp-predicate
      disable-locking
      process-queue-length)))

;;;----------------------------------------------------------------------------


(defun initialize-priority-queue 
       (priority-queue heap initial-contents element-type size key
	not-lessp-predicate disable-locking process-queue-length)
  
  (multiple-value-bind (heap length)
      (initialize-heap initial-contents size element-type key not-lessp-predicate heap)
    (let ((locking-style
            (cond (disable-locking nil)
                  ((and process-queue-length (plusp process-queue-length))
                   :QUEUED)
                  (t t))))
      (setf (priority-queue-heap priority-queue) heap
            (priority-queue-length priority-queue) length
            (priority-queue-sort-function priority-queue) not-lessp-predicate
            (priority-queue-key-function priority-queue) key
            (priority-queue-locking-style priority-queue) locking-style)
      (reset-priority-queue-lock priority-queue process-queue-length)))
  (values priority-queue))

(defun initialize-heap 
       (initial-contents size element-type key not-lessp-predicate &optional old-heap)
  (if initial-contents
      (let* ((length (length initial-contents))
	     (heap (or old-heap
		       ;; The heap is one-based.
		       (make-array (heap-actual-size size initial-contents 1)
                                   #-Explorer :adjustable #-Explorer t ; everything is adjustable on the LISPM
				   ;; this is redundant with `replace' below
				   ;; :initial-contents initial-contents
				   :element-type element-type))))
	
	;; The heap is one-based.
	(replace heap initial-contents :start1 1 :end1 (1+ length))
	(reorder-heap heap length key not-lessp-predicate)
	(values heap length))
      (values
	(or old-heap
	    ;; The heap is one-based.
	    (make-array (1+ size) :element-type element-type
                        #-Explorer :adjustable #-Explorer t ; everything is adjustable on the LISPM
                        ))
	0)))

(defun heap-actual-size (requested-size initial-contents &optional (offset 1))
  (max (+ (length initial-contents) offset) requested-size))

;;;----------------------------------------------------------------------------

(defun priority-queue-reset (priority-queue)
  "Resets the priority queue to an initial, empty state.  Ensures that it is
properly locked while emptying it."
  (without-interrupts
    (reset-priority-queue-lock priority-queue)
    (priority-queue-empty priority-queue)))

(defun reset-priority-queue-lock (priority-queue &optional process-queue-length)
  #-Explorer
  (declare (ignore priority-queue process-queue-length))
  #+Explorer
  (let ((locking-style (priority-queue-locking-style priority-queue)))
    (unwind-protect
        (progn
          ;; Temporarily disable locking
          (setf (priority-queue-locking-style priority-queue) nil)
          (ecase locking-style
            (:QUEUED
             ;; Extra paranoic checking for the case the locking-style has
             ;; changed since the last time we did this.
	     (cond ((typep (priority-queue-lock priority-queue) 'sys:process-queue)
		    (sys:reset-process-queue (priority-queue-lock priority-queue)))
                   (t
                    (assert process-queue-length nil
                            "the priority queue locking style changed to :QUEUED when it shouldn't have")
                    (setf (priority-queue-lock priority-queue)
                          (sys:make-process-queue
			    (symbol-name (newsym '|Priority Queue Locking Queue |))
                            process-queue-length)))))
            ((t)
             (setf (priority-queue-lock priority-queue) nil))
            ((nil) nil)))
      (setf (priority-queue-locking-style priority-queue) locking-style))))

;;;----------------------------------------------------------------------------

(defun print-priority-queue (priority-queue stream depth)
  (declare (ignore depth))
  #+Explorer
  (sys:printing-random-object (priority-queue stream)
    (princ (if (priority-queue-locking-style priority-queue) 
	       "Lockable-Priority-Queue"
	       "Priority-Queue")
	   stream)
    (write-char #\space stream)
    (princ (priority-queue-length priority-queue) stream))
  #-Explorer
  (format stream "#<~a ~d>"
	  (if (priority-queue-locking-style priority-queue) 
	      "Lockable-Priority-Queue"
	      "Priority-Queue")
	  (priority-queue-length priority-queue)))

;;;----------------------------------------------------------------------------

(defmacro with-priority-queue-locked ((priority-queue) &body body)
  "Insure exclusive access to `priority-queue' while `body' is executing.  This
function only works on Explorers; it's a no-op on other lisp platforms, but most
other lisp platforms don't have multiple threads, so it should be okay."
  #+Explorer
  `(sys:with-lock-if
     ((priority-queue-locking-style ,priority-queue)
      (priority-queue-lock ,priority-queue)
      :queued (eq (priority-queue-locking-style ,priority-queue) :QUEUED)
      :whostate "Priority Queue Lock")
     . ,body)
  #-Explorer
  (declare (ignore priority-queue))
  #-Explorer
  `(progn . ,body))

;;;----------------------------------------------------------------------------

;; Used inline below.
(proclaim '(inline priority-queue-empty-p))
(defun priority-queue-empty-p (priority-queue)
  "Returns T if the length of the priority queue is 0."
  (= (priority-queue-length priority-queue) 0))

(defun priority-queue-insert (element priority-queue)
  "Insert an element into a priority queue."
  (with-priority-queue-locked (priority-queue)
    (let ((length (incf (priority-queue-length priority-queue)))
          (heap (priority-queue-heap priority-queue)))
       ;; The heap is one-based (note <=).
      (when (<= (array-total-size heap) length)
        (adjust-array heap (round (* 1.5 (array-total-size heap)))))
      (setf (aref heap length) element)
      (siftup heap length
              (priority-queue-key-function priority-queue)
              (priority-queue-sort-function priority-queue))
      element)))

;;; Used inline below.
(proclaim '(inline priority-queue-pop))
(defun priority-queue-pop (priority-queue)
  "Pop an element from a priority-queue."
  (with-priority-queue-locked (priority-queue)
    (let ((length (decf (priority-queue-length priority-queue)))
          (heap (priority-queue-heap priority-queue)))
      (when (< length 0)
        (setf (priority-queue-length priority-queue) 0)
        (error "Attempt to pop empty priority queue"))
      (prog1
        (aref heap 1)
        (setf (aref heap 1) (aref heap (1+ length)))
        (siftdown heap 1 length
                  (priority-queue-key-function priority-queue)
                  (priority-queue-sort-function priority-queue))))))

(defun priority-queue-head (priority-queue)
  "Return the first element in the priority queue."
  (aref (priority-queue-heap priority-queue) 1))

(defun priority-queue-head-or-nil (priority-queue)
  "Return the first element in the priority queue or NIL if the queue is empty."
  (declare (inline priority-queue-empty-p))

  (with-priority-queue-locked (priority-queue)
    (and (not (priority-queue-empty-p priority-queue))
         (aref (priority-queue-heap priority-queue) 1))))

(defun priority-queue-find (item priority-queue &key key test)
  "Returns an priority-queue element matching `item.' This is just like the
Common Lisp `find' function."
  (with-priority-queue-locked (priority-queue)
    (let ((heap (priority-queue-heap priority-queue))
	  (end  (1+ (priority-queue-length priority-queue))))
      (if (null test)
	  (if (null key)
	      (find item heap :start 1 :end end)
	      (find item heap :key key :start 1 :end end))
	  (if (null key)
	      (find item heap :test test :start 1 :end end)
	      (find item heap :key key :test test :start 1 :end end))))))

(defun priority-queue-delete (element priority-queue)
  "Delete an element from the middle of a priority queue.  Element must be eql
to the queue element."
  (declare (inline priority-queue-pop))
  
  (with-priority-queue-locked (priority-queue)
    (let ((n nil)
          (heap (priority-queue-heap priority-queue)))
      ;; We use a loop here because it is faster than the `position' function.
      (loop named pos
	    for i from 1 to (1+ (priority-queue-length priority-queue)) do
	    (when (eql (aref heap i) element)
	      (setf n i)
	      (return-from pos)))
      
      (when n
        ;; delete element at index n, move 'hole' to top, then do a shiftdown
        (do (p)
            ((= 1 n))
          (setf p (ash n -1))
          (setf (aref heap n) (aref heap p))
          (setf n p))
        
        (priority-queue-pop priority-queue))
      
      (and n t))))

(defun priority-queue-empty (priority-queue)
  "Makes the the priority queue be empty."
  (clear-priority-queue priority-queue))

(defun clear-priority-queue (priority-queue)
  "Clear out a priority queue; leave it with no entries."
  (with-priority-queue-locked (priority-queue)
    (setf (priority-queue-length priority-queue) 0)))

;;; ============================================================================

(defmacro do-priority-queue ((var priority-queue) &body body)
  "Pops elements of the priority queue until it is empty.  Each element is bound
to `var.' This leaves the priority queue empty.  Uses `do,' so you can use
`return' to exit the loop early."
  (let ((pq (gentemp "PRIORITY-QUEUE")))
    `(let ((,pq ,priority-queue))
       (do () (nil)
	 (let ((,var (priority-queue-pop ,pq)))
	   . ,body)
	 (when (priority-queue-empty-p ,pq)
	   (return))))))

#+test
(defun test-do-priority-queue ()
  (let ((pq (make-priority-queue #'< :initial-contents '(3 6 5 1 4 2) :size 10)))
    (do-priority-queue (elt pq)
      (format t "~s " elt))))

;;;----------------------------------------------------------------------------

(defun priority-queue-map-in-priority-order (function priority-queue &rest extra-args)
  "Maps \\ar{function} over the elements of \\ar{priority-queue} in priority
order.  The arguments to \\ag{Function} are the queue element, the index and
then the \\ar{extra-args}.  Note that
\\fun{priority-queue-map-in-priority-order} is about thirty times slower than
\\fun{priority-queue-map-in-heap-order} and also conses."
  
  (flet ((pop-heap (heap length key sort)
	   "This is the guts of pop-priority-queue."
           (prog1
             (aref heap 1)
             (setf (aref heap 1) (aref heap length))
             (siftdown heap 1 (1- length) key sort))))
    (declare (inline pop-heap))
    
    (let* ((length (priority-queue-length priority-queue))
           (for-end (1+ length))
           (key (priority-queue-key-function priority-queue))
           (sort (priority-queue-sort-function priority-queue)))
      ;; Make a temporary copy of the heap.
      (let (temp)
        (unwind-protect
            (progn
              ;; Lock it while we are copying the heap to the temp array.
              (with-priority-queue-locked (priority-queue)
                (setf temp
		      (let ((a (make-array (1+ length))))
			(replace a (priority-queue-heap priority-queue))
			a)))
              (loop for n from 1 below for-end do
		    (apply function 
			   (pop-heap temp (- for-end n) key sort) n extra-args)))
          temp)))))

;;;----------------------------------------------------------------------------

(defun priority-queue-map-in-heap-order (function priority-queue &rest extra-args)
   "Apply `function' to each item in `priority-queue'; ignore values.
`Function's arguments are the queue element, the index and then the
`extra-args'.  NOTE: This does not process the elements in sorted order, but in
heap order, which is more efficient."
  (with-priority-queue-locked (priority-queue)
    (loop with heap = (priority-queue-heap priority-queue)
	  for index from 1 to (priority-queue-length priority-queue)
	  for elt = (aref heap index) do
	  (apply function elt index extra-args))))

(defun priority-queue-elements (priority-queue)
  "Returns a list of all the elements of the priority queue in heap order, not
in priority order.  The priority queue is unchanged."
  (with-priority-queue-locked (priority-queue)
    (coerce (subseq (priority-queue-heap priority-queue)
		    1 (1+ (priority-queue-length priority-queue)))
	    'list)))

;;;----------------------------------------------------------------------------

(defun print-priority-queue-elements
       (priority-queue &optional (stream *standard-output*))
  "Prints the elements of `priority-queue' to `stream,' which defaults to
*standard-output*.  Prints them in priority order."
  (format stream "~&")
  (priority-queue-map-in-priority-order
    #'(lambda (element index)
	(format stream "~&~2d. ~s" index element))
                      priority-queue))

;;;----------------------------------------------------------------------------

(defun reorder-priority-queue (priority-queue)
  "If your ordering predicate has changed, you can call this function to reorder
the elements of the priority queue and restore the heap property.  This function
is O\(n\), which is cheaper than sorting."
  (with-priority-queue-locked (priority-queue)
    (reorder-heap
      (priority-queue-heap priority-queue)
      (priority-queue-length priority-queue)
      (priority-queue-key-function priority-queue)
      (priority-queue-sort-function priority-queue))))

;;;----------------------------------------------------------------------------
;;; Heap functions
;;; Taken from programming pearls

(defmacro first-higher-priority-p (elt1 elt2 key-fn sort-fn)
  "Equivalent to \(funcall `sort-fn' \(funcall `key-fn' `elt1'\) \(funcall
`key-fn' `elt2'\)\) except that it avoids `funcall' \(which is expensive\) in
the special cases of `key-fn' = `identity' and `sort-fn' = `<=.' Note that this
uses three times as much space as the `funcall' code it is equivalent to \(36
microcode instructions versus 12\)."
  (let ((var1 (gensym "ARG"))
	(var2 (gensym "ARG")))
    `(if ,key-fn
	 (let ((,var1 (funcall ,key-fn ,elt1))
	       (,var2 (funcall ,key-fn ,elt2)))
	   (if (eq ,sort-fn #'<=)
	       (<= ,var1 ,var2)
	       (funcall ,sort-fn ,var1 ,var2)))
	 (if (eq ,sort-fn #'<=)
	     (<= ,elt1 ,elt2)
	     (funcall ,sort-fn ,elt1 ,elt2)))))

;;; These are assumed to be called with the entire structure locked.

(defun siftdown (heap i n key sort)
  "Move heap\(i\) downto the correct place."

  ;; Some error checking to prevent screwups during the reimplementation.
  (check-type heap array)
  (assert (not (typep heap 'priority-queue)) () 
	  "heap's type was `priority-queue', it should `array'")
  
  (let ((newelement (aref heap i)))
    (do (heapc
         ;; In lieu of (setf c (* i 2)) or (setf c (ash i 1)).
	 (c (+ i i) (+ i i)))
        (nil)
      
      (cond
        ;; either last element or out of bounds
        ((<= n c)
         (when (< n c)
           (return))
         ;; last element
         (setf heapc (aref heap c)))
        ;; c+1 is guaranteed to be in bounds. This code does less work if the
        ;; predicate returns T for elements of equal priority.
        ((not (first-higher-priority-p
		(setf heapc (aref heap c)) (aref heap (1+ c)) key sort))
         (setf heapc (aref heap (incf c)))))
      
      ;; This code does less work if the predicate returns T for elements of
      ;; equal priority.
      (when (first-higher-priority-p newelement heapc key sort)
	(return))
      
      (setf (aref heap i) heapc
	    i c))
    
    (setf (aref heap i) newelement)))

(defun siftup (heap n key sort)
  "Move heap(n) up to correct place"
  ;; i = some node; p = parent(i).  Move p into i and continue up the heap until
  ;; p is higher priority than the new element, in which case we stick the new
  ;; element into i.  This function does less work if the predicate returns T
  ;; for elements of equal priority.

  ;; Some error checking to prevent screwups during the reimplementation.
  (check-type heap array)
  (assert (not (typep heap 'priority-queue)) ()
	  "heap's type was `priority-queue', it should `array'")
  
  (let ((new-elt (aref heap n)))
    (do* ((i n p)
	  (p (ash i -1) (ash i -1)))
	 ((or (zerop p)
	      ;; This code could be made faster by caching the result of
	      ;; (funcall key new-elt)
	      (first-higher-priority-p (aref heap p)  new-elt key sort))
	  (setf (aref heap i) new-elt))
      (setf (aref heap i) (aref heap p)))))

(defun reorder-heap (heap length key sort)
  "Completely sorts the heap. Useful if your ordering predicate has changed."

  ;; Some error checking to prevent screwups during the reimplementation.
  (check-type heap array)
  (assert (not (typep heap 'priority-queue)) ()
	  "heap's type was `priority-queue', it should `array'")
  
  (do ((index (floor length 2) (- index 1)))
	((<= index 0))
      (siftdown heap index length key sort)))


;;; ============================================================================
;;; Testing code

#+test
(defun test-priority-queue ()
  (let ((pq (make-priority-queue #'< :size 10)))
    (priority-queue-insert 3 pq)
    (priority-queue-insert 2 pq)
    (priority-queue-insert 4 pq)
    (priority-queue-insert 1 pq)
    (format t "~&Length should be 4:  ~s~%" (priority-queue-length pq))
    (print-priority-queue pq *standard-output* t)
    (print-priority-queue-elements pq)
    ;; (priority-queue-map-in-heap-order #'(lambda (x y) (print x)) pq)
    (priority-queue-delete 2 pq)
    (format t "~&After deleting the number 2, the length should be 3:  ~s~%"
	    (priority-queue-length pq))
    (print-priority-queue-elements pq)
    (let ((elt (priority-queue-find 3 pq)))
      (format t "~&If it found the number 3, this should be it:  ~s~%" elt)
      (priority-queue-delete elt pq)
      (format t "~&And now it's deleted:~%")
      (print-priority-queue-elements pq)))
  (format t "~2&Priority Queues with a key~%")
  (let ((pq (make-priority-queue #'< :size 10 :key #'car)))
    (priority-queue-insert (list 3) pq)
    (priority-queue-insert (list 2) pq)
    (priority-queue-insert (list 4) pq)
    (priority-queue-insert (list 1) pq)
    (format t "~&Length should be 4:  ~s~%" (priority-queue-length pq))
    (print-priority-queue pq *standard-output* t)
    (print-priority-queue-elements pq)
    (priority-queue-delete (list 2) pq)
    (format t "~&After failing to delete the list containing 2,~%~
                 because it's not eql to the element, ~%~
                 the length should still be 4:  ~s~%"
	    (priority-queue-length pq))
    (print-priority-queue-elements pq)
    (let ((elt (priority-queue-find 3 pq :key #'car)))
      (format t "~&If it found the number 3, this should be it:  ~s~%" elt)
      (priority-queue-delete elt pq)
      (format t "~&And now it's deleted:~%")
      (print-priority-queue-elements pq)))
  )

;;; ============================================================================

(defmacro benchmark-priority-queue (name docstring init-form incf-form)
  "Reports on the efficiency of the priority queue.  This data structure is
often the inner loop of a system and therefore needs to be as efficient as
possible.  This function helps us discover the speed of the queue in terms of
how many pushes and pops per second it can do.  

This function varies the number of elements in the queue, from one to
`max-elements' and, for each such condition, does `num-pops' operations on the
queue.  It reports the number of operations per second."
  `(defun ,name (filename max-elements &optional num-pops) ,docstring
     (with-conditional-open-file (out filename :direction :output)
        (format out "~&# max  operations per second.~%")
        (format out "0 0~%") ; to make sure origin is graphed.
        (loop with q = (make-priority-queue #'>=
				            :key #'(lambda (x) (aref x 0))
				            :size (1+ max-elements))
	      for n from 2 to max-elements do
	      ;; initialize queue
	      (dotimes (i n)
	        (priority-queue-insert
                 (make-array 1 :initial-contents (list ,init-form)) q))
	      ;; do the timing
	      (let ((time (timeit (dotimes (i num-pops)
			            (let ((elt (priority-queue-pop q)))
				      (incf (aref elt 0) ,incf-form)
				      (priority-queue-insert elt q))))))
	        ;; report results
	        (format out "~3d ~20:@<~7,2f~>~%" n (float (/ num-pops time))))))))

(benchmark-priority-queue benchmark-priority-queue1
  "It pops the top element and increments the value by the size of the heap,
so that it will go to the bottom of the heap, and then re-inserts it.  Your
pattern of use may be very different."
  i
  max-elements)

(benchmark-priority-queue benchmark-priority-queue2
  "It pops the top element and increments the value by 1, so that it will go
past all the others at its level, and then re-inserts it.  Your pattern of use
may be very different."
  0
  1)

;;; ***************************************************************************
;;; EOF
