;;;; **************************************************************************
;;;; **************************************************************************
;;;; *                                                                        *
;;;; *                       PROCESS MONITOR PACKAGE                          *
;;;; *                                                                        *
;;;; **************************************************************************
;;;; **************************************************************************
;;;
;;; Written by: Matt Schmill
;;;             Department of Computer and Information Science
;;;             University of Massachusetts
;;;             Amherst, Massachusetts 01003.
;;;
;;;  04-16-95 File Created.  (Schmill)
;;;

(defpackage sysmon
  (:use clim clim-lisp))
(in-package :sysmon)

(export '(service-interrupt
          system-display-monitor
          monitor-system))

(defparameter *sysmon-cache* '())
(defparameter *sysmon-active-heap* '())
(defparameter +button-height+ 36)
;;
;;

(define-application-frame system-monitor ()
  ((interrupt-callback :initform NIL :initarg :interrupt-callback 
		       :accessor interrupt-callback)
   (rearm-interrupt :initform NIL :initarg :interrupt-rearm
                    :accessor interrupt-rearm)
   (sysid :initform NIL :initarg :sysid :accessor sysid)
   (refresh-process :initform NIL :initarg :refresh-process :accessor refresh-pid)
   (buttons :initform NIL :accessor sysmon-buttons))
  (:panes
   (display :application :scroll-bars :vertical :min-width 400
            :incremental-redisplay t :display-function 'display-sysmon-log
            #+ignore :end-of-page-action #+ignore :allow :end-of-line-action :allow)
   (buttons :application :scroll-bars :horizontal :min-width 400 :min-height +button-height+
            :end-of-line-action :allow :end-of-page-action :allow :height +button-height+
            :max-height +button-height+))
  (:layouts
   (default (vertically () display buttons))))

;;
;; here's the display-sysmon-log function, which is really just a shell
;;

(defmethod display-sysmon-log ((self system-monitor) stream)
  (updating-output (stream)
    (system-display-monitor (sysid self) self stream)))

(defgeneric system-display-monitor (system-id frame stream))

;;
;; here's the handle-interrupt method
;;

#+NO
(defgeneric service-interrupt (sysid))
#+NO
(defmethod handle-interrupt ((sysmon system-monitor))
  ;; interrupt the system
  (funcall (interrupt-callback sysmon))
  ;; handle the interrupt
  (service-interrupt (sysid sysmon))
  ;; rearm the system callback
  #+ignore (funcall (interrupt-rearm sysmon)))
#+NO
(defmethod resume ((sysmon system-monitor))
  (funcall (interrupt-rearm sysmon)))

;;
;; here's the exit method
;;

(defmethod exit-sysmon ((sysmon system-monitor))
  (setf *sysmon-active-heap* 
        (delete (rassoc sysmon *sysmon-active-heap* )
                *sysmon-active-heap*))
  (mp:process-kill (refresh-pid sysmon))
  (frame-exit sysmon))

;;
;; here's the start-system-monitor thingy
;;

(defun start-system-monitor (sysid start-options frame-options cache-key 
                                   refresh-when-fun refresh-reset-fun
                                   button-defs button-names)
  (let* ((new-frame (apply #'make-application-frame 'system-monitor
                           #+clim-2 :frame-manager
                           #+clim-2 (let ((port (find-port)))
                                      (typecase port
                                        (clim::frame-manager port)
                                        (t (clim::find-frame-manager :port port))))
                           :label (format nil "~A system monitor" sysid)
                           :sysid sysid
                           ;; :interrupt-callback interrupt-fun
                           ;; :interrupt-rearm reactivate-fun
                           frame-options))
         (gadstrm (get-frame-pane new-frame 'buttons))
         tmp)
    (format t "~S~%~S~%" start-options frame-options)
    (when new-frame
      (push (cons sysid new-frame) *sysmon-active-heap*)
      (when cache-key
        (push (cons cache-key new-frame) *sysmon-cache*))
      ;; catch funny monitor glitches
      (when (and (find :dwim *features*) (not (getf start-options :master)))
	(warn "DWIM is loaded, but no :MASTER has been supplied to start-system-monitor. ~%To use DWIM, supply a master frame."))
      (when (and (not (find :dwim *features*)) (getf start-options :master))
        (warn "DWIM has not been loaded, start-frame options ignored."))
      ;; start the frame up
      (if (and (getf start-options :master)
               (find :dwim *features*))
          (apply #'mp::process-run-function (format nil "~A monitor frame" sysid)
	         '(:run-reasons :enable)
		 (if (find-package :dwim)
                     (find-symbol 'start-frame :dwim)
                   'run-frame-top-level)
	         new-frame
	         start-options)
        (mp::process-run-function (format nil "~A monitor frame" sysid)
                                  '(:run-reasons :enable)
                                  #'run-frame-top-level new-frame))
      ;; draw the buttons
      (loop for bd in (append button-defs (list #'(lambda (button) 
						    (exit-sysmon (pane-frame button)))))
            for bn in (append button-names '("Exit")) do
	    (case bd
              (:space
	       (stream-set-cursor-position gadstrm (+ (stream-cursor-position gadstrm) +button-height+) 0))
	      (t
	       (multiple-value-bind (x y)
		   (stream-cursor-position gadstrm)
		 (declare (ignore y))
		 (stream-set-cursor-position gadstrm x 0))
	       (with-output-as-gadget (gadstrm)
		 (setf tmp (make-pane 'push-button :label (princ-to-string bn)
				      :activate-callback bd)))
	       (push tmp (sysmon-buttons new-frame)))))
      ;; now start the refresh process
      #+LISPWORKS
      (setf (refresh-pid new-frame)
            (mp::process-run-function (format nil "~A sysmon refresh process" sysid)
                                      '(:run-reasons :enable)
                                      (COMPILE NIL
                                        `(LAMBDA (SYSID)
					   (DECLARE (IGNORE SYSID))
					   (LOOP WHILE T DO
						 (PROGN
						   (MP:PROCESS-WAIT "Background..." ,REFRESH-WHEN-FUN)
						   ;(REDISPLAY-FRAME-PANES ,NEW-FRAME)
                                                   (FUNCALL ,REFRESH-RESET-FUN)
						   (redisplay-frame-pane ,new-frame 'display :force-p t)))))
                                      #+OLD (make-refresh-function new-frame 'display
                                                             refresh-when-fun refresh-reset-fun)
                                      sysid))
      #-LISPWORKS
      (error "Dig out the code, pal, it's written for LispWorks.")
      new-frame)))

;;
;; nice macro for starting the system monitor with a cache option
;;

(defmacro build-lambdas (forms)
  `(list ,@(mapcar #'(lambda (f)
                       (if (consp f)
		           `(function (lambda ,(second f)
				        ,@(rest (rest f))))
                           f))
                   forms)))

(defmacro build-names (forms)
  `(list ,@(mapcar #'(lambda (f)
                       (if (consp f)
                           (first f)
                           f))
                   forms)))

(defmacro monitor-system ((system-id &rest sm-options)
                          (&rest frame-options)
			  refresh-when-form refresh-reset-form
                          &rest button-defs)
  `(let ((mod-button-defs (build-lambdas ,button-defs))
         (names (build-names ,button-defs)))
     (if (and (assoc ,system-id sysmon::*sysmon-cache*)
              (getf ',sm-options :sysmon-cache))
         (clim::window-expose (cdr (assoc ,system-id sysmon::*sysmon-cache*)))
       (let ((cache-p (getf ',sm-options :sysmon-cache))
             (nopts (list ,@sm-options)))
         (remf nopts :sysmon-cache)
         (sysmon::start-system-monitor ,system-id nopts ',frame-options cache-p
				       (function (lambda NIL ,refresh-when-form))
				       (function (lambda NIL ,refresh-reset-form))
				       mod-button-defs
                                       names)))))

#+HOLD
(sysmon:monitor-system (:demo)
                ()
                (do-s 'this)
		(do-s 'that)
                (biddle (button) (frame-exit (pane-frame button)))
                (wank (button) (refresh-frame-panes (pane-frame button))))
